/*

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    atomic integer operations dispatch

*/

#ifndef _GCT_ATOMIC_H_
#define _GCT_ATOMIC_H_

/* Try to guess atomic mechanism if none already defined */

#if !(defined(GCT_CONFIG_ATOMIC_PTHREAD_SPIN) || defined(GCT_CONFIG_ATOMIC_PTHREAD_MUTEX) || defined(GCT_CONFIG_ATOMIC_GCC) || defined(GCT_CONFIG_ATOMIC_NONE) || defined(GCT_CONFIG_ATOMIC_OTHER))

# warning No atomic operation mechanism defined, guessing ...

# if (((__GNUC__ == 4) && (__GNUC_MINOR__ >= 1)) || __GNUC__ > 4)
#  define GCT_CONFIG_ATOMIC_GCC
#  warning Using gcc atomic builtins
  
# elif defined(_REENTRANT) || defined(_THREAD_SAFE)
#  if (_XOPEN_SOURCE >= 600)
#   define GCT_CONFIG_ATOMIC_PTHREAD_SPIN
#   warning Using pthread spinlock locking
#  else
#   define GCT_CONFIG_ATOMIC_PTHREAD_MUTEX
#   warning Using pthread mutex locking
#  endif

# else
#  error Unable to select an atomic operation mechanism, please define one using GCT_CONFIG_ATOMIC_* macros
# endif

#endif

/* Include selected atomic mechanism header */

#if defined(GCT_CONFIG_ATOMIC_PTHREAD_SPIN) || defined (GCT_CONFIG_ATOMIC_PTHREAD_MUTEX)
# include <gct/_atomic_pthread.h>

#elif defined(GCT_CONFIG_ATOMIC_GCC)
# include <gct/_atomic_gcc.h>

#elif defined(GCT_CONFIG_ATOMIC_NONE)
# include <gct/_atomic_none.h>

#elif defined(GCT_CONFIG_ATOMIC_OTHER)
 /* Using user predefined atomic mechanism */

#else
# error No atomic operation mechanism defined
#endif

_GCT_INTERNAL void _gct_atomic_init(_gct_atomic_t *a);
_GCT_INTERNAL void _gct_atomic_destroy(_gct_atomic_t *a);
_GCT_INTERNAL void _gct_atomic_set(_gct_atomic_t *a, _gct_atomic_int_t v);
_GCT_INTERNAL _gct_atomic_int_t _gct_atomic_get(_gct_atomic_t *a);
_GCT_INTERNAL gct_bool_t _gct_atomic_inc(_gct_atomic_t *a);
_GCT_INTERNAL gct_bool_t _gct_atomic_dec(_gct_atomic_t *a);
_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test(_gct_atomic_t *a, _gct_atomic_int_t n);
_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test_set(_gct_atomic_t *a, _gct_atomic_int_t n);
_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test_clr(_gct_atomic_t *a, _gct_atomic_int_t n);
_GCT_INTERNAL void _gct_atomic_bit_clr(_gct_atomic_t *a, _gct_atomic_int_t n);
_GCT_INTERNAL void _gct_atomic_bit_set(_gct_atomic_t *a, _gct_atomic_int_t n);

#endif

