/*

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    dummy (non)atomic integer operations

*/

#ifndef _GCT_ATOMIC_GCC_H_
#define _GCT_ATOMIC_GCC_H_

#if !(((__GNUC__ == 4) && (__GNUC_MINOR__ >= 1)) || __GNUC__ > 4)
# error GCC atomic operations not supported by the compiler, gcc >= 4.1.0 required
#endif

#include <gct/_platform.h>

typedef _GCT_INT     _gct_atomic_int_t;

typedef struct _gct_atomic_s
{
  volatile _gct_atomic_int_t val;
} _gct_atomic_t;

# define _GCT_ATOMIC_INITIALIZER(v)     { v }

_GCT_INTERNAL void _gct_atomic_init(_gct_atomic_t *a)
{
}

_GCT_INTERNAL void _gct_atomic_destroy(_gct_atomic_t *a)
{
}

_GCT_INTERNAL void _gct_atomic_set(_gct_atomic_t *a, _gct_atomic_int_t v)
{
  a->val = v;
}

_GCT_INTERNAL _gct_atomic_int_t _gct_atomic_get(_gct_atomic_t *a)
{
  return a->val;
}

_GCT_INTERNAL gct_bool_t _gct_atomic_inc(_gct_atomic_t *a)
{
  return __sync_add_and_fetch(&a->val, 1) != 0;
}

_GCT_INTERNAL gct_bool_t _gct_atomic_dec(_gct_atomic_t *a)
{
  return __sync_sub_and_fetch(&a->val, 1) != 0;
}

_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  return (a->val & (1 << n)) != 0;
}

_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test_set(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     mask = (1 << n);

  return (__sync_fetch_and_or(&a->val, mask) & mask) != 0;
}

_GCT_INTERNAL gct_bool_t _gct_atomic_bit_test_clr(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     mask = (1 << n);

  return (__sync_fetch_and_and(&a->val, ~mask) & mask) != 0;
}

_GCT_INTERNAL void _gct_atomic_bit_clr(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     mask = (1 << n);

  __sync_fetch_and_and(&a->val, ~mask);
}

_GCT_INTERNAL void _gct_atomic_bit_set(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     mask = (1 << n);

  __sync_fetch_and_or(&a->val, mask);
}

#endif

