/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_BITOPS_H_
#define _GCT_BITOPS_H_

#include <gct/_platform.h>

# define _GCT_BIT_TEST(type, data, bit)       ((data) & ((type)1 << (bit)) ? 1 : 0)
# define _GCT_BIT_SET(type, data, bit)        ((data) |= ((type)1 << (bit)))
# define _GCT_BIT_CLR(type, data, bit)        ((data) &= ~((type)1 << (bit)))

/* backslash-region-begin */
# define _GCT_BIT_SETVALUE(type, data, bit, value)			\
  {									\
    if (value)								\
      _GCT_BIT_SET(type, data, bit);					\
    else								\
      _GCT_BIT_CLR(type, data, bit);					\
  }									\
/* backslash-region-end */

#if __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)

/* backslash-region-begin */
# define _GCT_POPCOUNT(n)						\
  (__builtin_types_compatible_p(typeof(n), _GCT_ULONG) ? __builtin_popcountl(n) : \
  __builtin_types_compatible_p(typeof(n), _GCT_ULONGLONG) ? __builtin_popcountll(n) : \
  __builtin_popcount(n))						\
/* backslash-region-end */

/* backslash-region-begin */
# define _GCT_BIT_FFS(n)						\
  (__builtin_types_compatible_p(typeof(n), _GCT_ULONG) ? __builtin_ffsl(n) : \
  __builtin_types_compatible_p(typeof(n), _GCT_ULONGLONG) ? __builtin_ffsll(n) : \
  __builtin_ffs(n))							\
/* backslash-region-end */

/* backslash-region-begin */
# define _GCT_BIT_FFSR(n)						\
({									\
  typeof(n) _gct_n = (n);						\
									\
  _gct_n ? (__builtin_types_compatible_p(typeof(_gct_n), _GCT_ULONG) ? sizeof(_GCT_ULONG) * 8 - __builtin_clzl(_gct_n) : \
  __builtin_types_compatible_p(typeof(_gct_n), _GCT_ULONGLONG) ? sizeof(_GCT_ULONGLONG) * 8 - __builtin_clzll(_gct_n) : \
  sizeof(_GCT_UINT) * 8 - __builtin_clz(_gct_n)) : 0;			\
})									\
/* backslash-region-end */

#elif __GNUC__

/* backslash-region-begin */
# define _GCT_POPCOUNT(n)						\
({									\
  uint_fast8_t        _gct_x, _gct_count = 0;				\
  typeof(n)     _gct_w = n;						\
									\
  for (_gct_x = 0; _gct_x < sizeof(_gct_w) * 8; _gct_x++)		\
    if (_GCT_BIT_TEST(typeof(n), _gct_w, _gct_x))			\
      _gct_count++;							\
									\
  _gct_count;								\
})									\
/* backslash-region-end */


/* backslash-region-begin */
# define _GCT_BIT_FFS(n)						\
({									\
  uint_fast8_t        _gct_x, _gct_pos = 0;				\
  typeof(n)     _gct_w = n;						\
									\
  for (_gct_x = 0; _gct_x < sizeof(_gct_w) * 8; _gct_x++)		\
    if (_GCT_BIT_TEST(typeof(n), _gct_w, _gct_x))			\
      {									\
        _gct_pos = _gct_x + 1;						\
        break;								\
      }									\
									\
  _gct_pos;								\
})									\
/* backslash-region-end */


/* backslash-region-begin */
# define _GCT_BIT_FFSR(n)						\
({									\
  int_fast8_t   _gct_x;							\
  uint_fast8_t  _gct_pos = 0;						\
  typeof(n)     _gct_w = n;						\
									\
  for (_gct_x = sizeof(_gct_w) * 8 - 1; _gct_x >= 0; _gct_x--)		\
    if (_GCT_BIT_TEST(typeof(n), _gct_w, _gct_x))			\
      {									\
        _gct_pos = _gct_x + 1;						\
        break;								\
      }									\
									\
  _gct_pos;								\
})									\
/* backslash-region-end */

#else

_GCT_INTERNAL uint_fast8_t
_GCT_POPCOUNT(uintmax_t _gct_word)
{
  uint_fast8_t        _gct_x, _gct_count = 0;

  for (_gct_x = 0; _gct_x < sizeof(_gct_word) * 8; _gct_x++)
    if (_GCT_BIT_TEST(uintmax_t, _gct_word, _gct_x))
      _gct_count++;

  return _gct_count;
}

_GCT_INTERNAL uint_fast8_t
_GCT_BIT_FFS(uintmax_t _gct_word)
{
  uint_fast8_t        _gct_x, _gct_pos = 0;

  for (_gct_x = 0; _gct_x < sizeof(_gct_word) * 8; _gct_x++)
    if (_GCT_BIT_TEST(uintmax_t, _gct_word, _gct_x))
      {
        _gct_pos = _gct_x + 1;
        break;
      }

  return _gct_pos;
}

_GCT_INTERNAL uint_fast8_t
_GCT_BIT_FFSR(uintmax_t _gct_word)
{
  int_fast8_t   _gct_x;
  uint_fast8_t  _gct_pos = 0;

  for (_gct_x = sizeof(_gct_word) * 8 - 1; _gct_x >= 0; _gct_x--)
    if (_GCT_BIT_TEST(uintmax_t, _gct_word, _gct_x))
      {
        _gct_pos = _gct_x + 1;
        break;
      }

  return _gct_pos;
}

#endif

#endif

