/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Container main access functions stuff

*/

#ifndef _GCT_CONT_ACCESS_H_
#define _GCT_CONT_ACCESS_H_

#include <stdarg.h>

#include <gct/_platform.h>
#include <gct/_cont_lock.h>
#include <gct/_cont_refcount.h>
#include <gct/_cont_foreach.h>

/***********************************************************************
 *      container access functions prototypes
 */

/* Itialize an item and mark it as not being part of a container. */
/* backslash-block */
#define GCT_CONT_ORPHAN_PROTOTYPE(name, prefix)                    
_GCT_UNUSED void                                                
prefix##_orphan(name##_item_t item)

/* Check if an item is currently part of a container */
/* backslash-block */
#define GCT_CONT_ISORPHAN_PROTOTYPE(name, prefix)                  
_GCT_UNUSED gct_bool_t                                         
prefix##_isorphan(name##_item_t item)

/* Check for null index */
/* backslash-block */
#define GCT_CONT_ISNULL_PROTOTYPE(name, prefix)                    
_GCT_UNUSED gct_bool_t                                                     
prefix##_isnull (name##_index_t index)

/* Check index validity */
/* backslash-block */
#define GCT_CONT_ISVALID_PROTOTYPE(name, prefix)                   
_GCT_UNUSED gct_bool_t                                                     
prefix##_isvalid(name##_root_t *root, name##_index_t index)

/* Return container empty state */
/* backslash-block */
#define GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)                   
_GCT_UNUSED gct_bool_t                                                     
prefix##_isempty(name##_root_t *root)

/* Return container full state */
/* backslash-block */
#define GCT_CONT_ISFULL_PROTOTYPE(name, prefix)                   
_GCT_UNUSED gct_bool_t                                                     
prefix##_isfull(name##_root_t *root)

/* Get the item at the specified index. */
/* backslash-block */
#define GCT_CONT_GET_PROTOTYPE(name, prefix)                       
_GCT_UNUSED name##_item_t                                                   
prefix##_get    (name##_root_t *root, name##_index_t index)

/* Get the pointer to item at the specified index. */
/* backslash-block */
#define GCT_CONT_GETPTR_PROTOTYPE(name, prefix)                    
_GCT_UNUSED name##_itembase_t *                                             
prefix##_getptr (name##_root_t *root, name##_index_t index)

/* Get item index from pointer to item. */
/* backslash-block */
#define GCT_CONT_GETINDEX_PROTOTYPE(name, prefix)                  
_GCT_UNUSED name##_index_t                                                  
prefix##_getindex(name##_root_t *root, name##_itembase_t *ptr)

/* Allocate a new unused item in the container and return its index */
/* backslash-block */
#define GCT_CONT_ALLOC_PROTOTYPE(name, prefix)                     
_GCT_UNUSED name##_index_t                                                  
prefix##_alloc (name##_root_t *root, ...)

/* Set an item and replace the old item at the specified index */
/* backslash-block */
#define GCT_CONT_SET_PROTOTYPE(name, prefix)                       
_GCT_UNUSED gct_err_t                                                    
prefix##_set    (name##_root_t *root,                           
                 name##_index_t index,                          
                 name##_item_t item)

/* Return the index of the next item */
/* backslash-block */
#define GCT_CONT_NEXT_PROTOTYPE(name, prefix)                      
_GCT_UNUSED name##_index_t                                                  
prefix##_next   (name##_root_t *root, name##_index_t index)

/* Return the index of the previous item */
/* backslash-block */
#define GCT_CONT_PREV_PROTOTYPE(name, prefix)                      
_GCT_UNUSED name##_index_t                                                  
prefix##_prev   (name##_root_t *root, name##_index_t index)

/* Return the index of the first item in the container */
/* backslash-block */
#define GCT_CONT_HEAD_PROTOTYPE(name, prefix)                      
_GCT_UNUSED name##_index_t                                                  
prefix##_head   (name##_root_t *root)

/* Return the index of the last item in the container */
/* backslash-block */
#define GCT_CONT_TAIL_PROTOTYPE(name, prefix)                      
_GCT_UNUSED name##_index_t                                                  
prefix##_tail   (name##_root_t *root)

/* Return the current items count in the container */
/* backslash-block */
#define GCT_CONT_COUNT_PROTOTYPE(name, prefix)                     
_GCT_UNUSED size_t                                                          
prefix##_count  (name##_root_t *root)

/* Return the maximum items count in the container */
/* backslash-block */
#define GCT_CONT_SIZE_PROTOTYPE(name, prefix)                      
_GCT_UNUSED size_t                                                          
prefix##_size(name##_root_t *root)

/* Change the maximum items count (realloc) */
/* backslash-block */
#define GCT_CONT_RESIZE_PROTOTYPE(name, prefix)                    
_GCT_UNUSED gct_err_t                                                    
prefix##_resize(name##_root_t *root, size_t size)

/* Insert an item before an other item in the container */
/* backslash-block */
#define GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix)                
_GCT_UNUSED size_t                                                          
prefix##_insert_pre(name##_root_t *root, name##_index_t index,  
                    name##_item_t item)

/* Insert an item next to an other item in the container */
/* backslash-block */
#define GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix)               
_GCT_UNUSED size_t                                                          
prefix##_insert_post(name##_root_t *root, name##_index_t index, 
                     name##_item_t item)

/* Delete the item at the given index in the container */
/* backslash-block */
#define GCT_CONT_REMOVE_PROTOTYPE(name, prefix)                    
_GCT_UNUSED gct_err_t                                                    
prefix##_remove (name##_root_t *root, name##_index_t index)

/* Insert an item before the container first item */
/* backslash-block */
#define GCT_CONT_PUSH_PROTOTYPE(name, prefix)                      
_GCT_UNUSED size_t                                                          
prefix##_push   (name##_root_t *root, name##_item_t item)

/* Insert an item after the container last item */
/* backslash-block */
#define GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)                  
_GCT_UNUSED size_t                                                          
prefix##_pushback(name##_root_t *root, name##_item_t item)

/* Remove and get the container first item */
/* backslash-block */
#define GCT_CONT_POP_PROTOTYPE(name, prefix)                       
_GCT_UNUSED name##_item_t                                                   
prefix##_pop    (name##_root_t  *root)

/* Remove and get the container last item */
/* backslash-block */
#define GCT_CONT_POPBACK_PROTOTYPE(name, prefix)                   
_GCT_UNUSED name##_item_t                                                   
prefix##_popback(name##_root_t  *root)

/* Insert several items before the container first item */
/* backslash-block */
#define GCT_CONT_PUSH_ARRAY_PROTOTYPE(name, prefix)                
_GCT_UNUSED size_t                                                          
prefix##_push_array     (name##_root_t *root,                   
                         name##_item_t *item,                   
                         size_t size)

/* Insert several items after the container last item */
/* backslash-block */
#define GCT_CONT_PUSHBACK_ARRAY_PROTOTYPE(name, prefix)            
_GCT_UNUSED size_t                                                          
prefix##_pushback_array (name##_root_t *root,                   
                         name##_item_t *item,                   
                         size_t size)

/* Remove several items from the _cont_access.head and copy them to an array */
/* backslash-block */
#define GCT_CONT_POP_ARRAY_PROTOTYPE(name, prefix)                 
_GCT_UNUSED size_t                                                          
prefix##_pop_array      (name##_root_t  *root,                  
                         name##_item_t *item,                   
                         size_t size)

/* Remove several items from the container tail and copy them to an array */
/* backslash-block */
#define GCT_CONT_POPBACK_ARRAY_PROTOTYPE(name, prefix)             
_GCT_UNUSED size_t                                                          
prefix##_popback_array  (name##_root_t  *root,                  
                         name##_item_t *item,                   
                         size_t size)

/* Container iterator function to be used with prefix##_foreach() function */
/* backslash-block */
#define GCT_CONTAINER_ITERATOR(name, fcn)                           
intptr_t (fcn) (name##_item_t item, va_list ap)

/* Iterate over the whole container by calling an iterator function.
   Stop iteration if the iterator function return non zero. */
/* backslash-block */
#define GCT_CONT_FOREACH_PROTOTYPE(name, prefix)                   
_GCT_UNUSED intptr_t                                                        
prefix##_foreach        (name##_root_t *root,                   
                         GCT_CONTAINER_ITERATOR(name, * const fcn), ...)

/* Iterate over the whole container by calling an iterator function.
   Stop iteration if the iterator function return non zero. */
/* backslash-block */
#define GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)           
_GCT_UNUSED intptr_t                                                        
prefix##_foreach_reverse(name##_root_t *root,                   
                         GCT_CONTAINER_ITERATOR(name, * const fcn), ...)

/* Remove all items from the container */
/* backslash-block */
#define GCT_CONT_CLEAR_PROTOTYPE(name, prefix)                     
_GCT_UNUSED void                                                            
prefix##_clear          (name##_root_t *root)

/* Init the container root */
/* backslash-block */
#define GCT_CONT_INIT_PROTOTYPE(name, prefix)                      
_GCT_UNUSED gct_err_t                                                    
prefix##_init           (name##_root_t *root)

/* Destroy the container root */
/* backslash-block */
#define GCT_CONT_DESTROY_PROTOTYPE(name, prefix)                   
_GCT_UNUSED void                                                            
prefix##_destroy        (name##_root_t *root)

/* Check container root consitency */
/* backslash-block */
#define GCT_CONT_CHECK_PROTOTYPE(name, prefix)                     
_GCT_UNUSED gct_err_t                                                    
prefix##_check          (name##_root_t *root)



/***********************************************************************
 *      container functions prototypes macro
 */

/* backslash-region-begin */
#define GCT_CONTAINER_PROTOTYPES(name, attr, prefix)

attr GCT_CONT_ISORPHAN_PROTOTYPE(name, prefix);
attr GCT_CONT_ORPHAN_PROTOTYPE(name, prefix);
attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix);
attr GCT_CONT_ISVALID_PROTOTYPE(name, prefix);
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix);
attr GCT_CONT_GET_PROTOTYPE(name, prefix);
attr GCT_CONT_GETPTR_PROTOTYPE(name, prefix);
attr GCT_CONT_GETINDEX_PROTOTYPE(name, prefix);
attr GCT_CONT_ALLOC_PROTOTYPE(name, prefix);
attr GCT_CONT_SET_PROTOTYPE(name, prefix);
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix);
attr GCT_CONT_PREV_PROTOTYPE(name, prefix);
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix);
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix);
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix);
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix);
attr GCT_CONT_RESIZE_PROTOTYPE(name, prefix);
attr GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix);
attr GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix);
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix);
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix);
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix);
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix);
attr GCT_CONT_POP_PROTOTYPE(name, prefix);
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix);
attr GCT_CONT_PUSH_ARRAY_PROTOTYPE(name, prefix);
attr GCT_CONT_PUSHBACK_ARRAY_PROTOTYPE(name, prefix);
attr GCT_CONT_POP_ARRAY_PROTOTYPE(name, prefix);
attr GCT_CONT_POPBACK_ARRAY_PROTOTYPE(name, prefix);
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix);
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix);
attr GCT_CONT_WRLOCK_PROTOTYPE(name, prefix);
attr GCT_CONT_RDLOCK_PROTOTYPE(name, prefix);
attr GCT_CONT_UNLOCK_PROTOTYPE(name, prefix);
attr GCT_CONT_INIT_PROTOTYPE(name, prefix);
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix);
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix);
/* backslash-region-end */

/***********************************************************************
 *      container types and functions defintions macro
 */

/* Define type associated with a container */
/* backslash-block */
#define         GCT_CONTAINER_TYPES(name, algo, type, ...)
  typedef _gct_empty_t _gct_lock_GCT_CONTAINER_LOCK_##name##_type_t;
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _TYPES)(name, type, __VA_ARGS__)
  _GCT_LOCK_EMPTY_LOCK_FCNS(_gct_lock_GCT_CONTAINER_LOCK_##name)
  _GCT_CONT_OBJECT_EMPTY_FCNS(name, GCT_CONTAINER_OBJ_##name)
  extern name##_item_t name##_item; /* fake object for custom_expr to be always valid for typeof() */

/* backslash-block */
#define         _GCT_CONTAINER_ENTRY_TYPE(algo)
  _GCT_CPP_CONCAT(_gct_, algo, _entry_t)

/* backslash-block */
#define         GCT_CONTAINER_ANCHOR(algo, declname)
  _GCT_CPP_CONCAT(_gct_, algo, _entry_t) declname;

/* define all container access functions */

/* backslash-block */
#define         GCT_CONTAINER_NOLOCK_FCNS(name, algo, attr, prefix, ...)
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FCNS)(attr, name, prefix, NOLOCK, __VA_ARGS__)


/* backslash-block */
#define         GCT_CONTAINER_LOCK_FCNS(name, algo, attr, prefix, lockname, ...)
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FCNS)(attr, name, prefix, lockname, __VA_ARGS__)


/* backslash-block */
#define         GCT_CONTAINER_FCNS(name, algo, attr, prefix, ...)
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FCNS)(attr, name, prefix, GCT_CONTAINER_LOCK_##name, __VA_ARGS__)
  _GCT_CONT_LOCK_FCNS(attr, name, prefix, GCT_CONTAINER_LOCK_##name)


/* Default container object intializer */
/* backslash-block */
#define         GCT_CONTAINER_ROOT_INITIALIZER(name, algo) 
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _ROOT_INITIALIZER)(name, GCT_CONTAINER_LOCK_##name)


/* backslash-block */
#define         GCT_CONTAINER_DECLARATOR(name, algo, symbol) 
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _ROOT_DECLARATOR)(name, GCT_CONTAINER_LOCK_##name, symbol)


/* backslash-block */
#define         GCT_CONTAINER_FEATURE(algo, test)
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _##test)

#endif

