/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_CONT_ARRAY_H_
#define _GCT_CONT_ARRAY_H_

/* backslash-region-begin */
#define _GCT_CONT_ARRAYS_FCNS(attr, name, prefix, lockname, algo)

attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)
{
  return index == name##_null_index;
}

attr GCT_CONT_ISVALID_PROTOTYPE(name, prefix)
{
  return ((index >= 0) && (index < root->count));
}

_GCT_NONNULL(1)
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)
{
  return root->count == 0;
}

_GCT_NONNULL(1)
attr GCT_CONT_GET_PROTOTYPE(name, prefix)
{
  name##_item_t         item;

  _GCT_ASSERT((index >= 0) && (index < root->count));

  item = root->data[index];

  _GCT_CONT_REFNEW(name, item);

  return item;
}

_GCT_NONNULL(1)
attr GCT_CONT_GETPTR_PROTOTYPE(name, prefix)
{
  return root->data + index;
}

_GCT_NONNULL(1)
attr GCT_CONT_GETINDEX_PROTOTYPE(name, prefix)
{
  _gct_sindex_t      i = (ptr - root->data);

  _GCT_ASSERT(i >= 0);
  _GCT_ASSERT(i < _GCT_CONT_##algo##_SIZE(name, root));
  _GCT_ASSERT(i < root->count);

  return i;
}

_GCT_NONNULL(1)
attr GCT_CONT_SET_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  _GCT_CONT_REFDROP(name, root->data[index]);
  root->data[index] = _GCT_CONT_REFNEW(name, item);

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return 0;
}

_GCT_NONNULL(1)
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)
{
  name##_index_t                res;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  res = index < root->count - 1
    ? index + 1
    : name##_null_index;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_PREV_PROTOTYPE(name, prefix)
{
  name##_index_t                res;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  res = index > 0
    ? index - 1
    : name##_null_index;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix)
{
  name##_index_t                res;

  res = root->count ? 0 : name##_null_index;

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix)
{
  name##_index_t                res;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  res = root->count
    ? root->count - 1
    : name##_null_index;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)
{
  return root->count;
}

_GCT_NONNULL(1)
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)
{
  return _GCT_CONT_##algo##_SIZE(name, root);
}

attr GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix)
{
  size_t        res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  if (_gct_cont_array_##name##_extend(root, 1))
    {
      _GCT_CONT_REFNEW(name, item);
      memmove(root->data + index + 1,
              root->data + index,
              (root->count - index) * sizeof(name##_item_t));
      root->data[index] = item;
      root->count++;
      res++;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

attr GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix)
{
  size_t        res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  if (_gct_cont_array_##name##_extend(root, 1))
    {
      _GCT_CONT_REFNEW(name, item);
      memmove(root->data + index + 2,
              root->data + index + 1,
              (root->count - index - 1) * sizeof(name##_item_t));
      root->data[index + 1] = item;
      root->count++;
      res++;
     }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  for (i = 0; i < root->count; i++)
    _GCT_CONT_REFDROP(name, root->data[i]);

  root->count = 0;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);
}

_GCT_NONNULL(1)
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  _GCT_ASSERT((index >= 0) && (index < root->count));

  _GCT_CONT_REFDROP(name, root->data[index]);

  memmove(root->data + index,
          root->data + index + 1,
          (root->count - index - 1) * sizeof(name##_item_t));

  root->count--;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return 0;
}

_GCT_NONNULL(1)
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)
{
  size_t        res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (_gct_cont_array_##name##_extend(root, 1))
    {
      memmove(root->data + 1, root->data,
              root->count * sizeof(name##_item_t));
      root->data[0] = item;
      root->count++;
      res++;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)
{
  size_t        res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (_gct_cont_array_##name##_extend(root, 1))
    {
      root->data[root->count++] = _GCT_CONT_REFNEW(name, item);
      res++;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_ALLOC_PROTOTYPE(name, prefix)
{
  name##_index_t        res = name##_null_index;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (_gct_cont_array_##name##_extend(root, 1))
    res = root->count++;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_POP_PROTOTYPE(name, prefix)
{
  name##_item_t res = name##_null_item;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (root->count > 0)
    {
      res = root->data[0];
      memmove(root->data, root->data + 1,
              (root->count - 1) * sizeof(name##_item_t));
      root->count--;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix)
{
  name##_item_t res = name##_null_item;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (root->count > 0)
    res = root->data[--root->count];

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_PUSH_ARRAY_PROTOTYPE(name, prefix)
{
  _gct_index_t     i, j;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  i = _gct_cont_array_##name##_extend(root, size);

  memmove(root->data + i, root->data,
          root->count * sizeof(name##_item_t));
  for (j = 0; j < i; j++)
    root->data[i - j - 1] = _GCT_CONT_REFNEW(name, item[j]);

  root->count += i;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return i;
}

_GCT_NONNULL(1)
attr GCT_CONT_PUSHBACK_ARRAY_PROTOTYPE(name, prefix)
{
  _gct_index_t     i, j;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  i = _gct_cont_array_##name##_extend(root, size);

  for (j = 0; j < i; j++)
    root->data[root->count + j] = _GCT_CONT_REFNEW(name, item[j]);

  root->count += i;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return i;
}

_GCT_NONNULL(1)
attr GCT_CONT_POP_ARRAY_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  i = size > root->count ? root->count : size;

  memcpy(item,
         root->data,
         i * sizeof(name##_item_t));
  memmove(root->data,
          root->data + i,
          (root->count - i) * sizeof(name##_item_t));

  root->count -= i;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return i;
}

_GCT_NONNULL(1)
attr GCT_CONT_POPBACK_ARRAY_PROTOTYPE(name, prefix)
{
  _gct_index_t     i, j;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  i = size > root->count ? root->count : size;

  for (j = 0; j < i; j++)
    item[j] = root->data[i - j - 1];
  memmove(root->data,
          root->data + i,
          i * sizeof(name##_item_t));

  root->count -= i;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return i;
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;
  intptr_t      res = 0;
  va_list       ap;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = 0; !res && i < root->count; i++)
    {
      va_start(ap, fcn);
      res = fcn(root->data[i], ap);
      va_end(ap);
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;
  intptr_t      res = 0;
  va_list       ap;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = root->count; !res && i > 0; i--)
    {
      va_start(ap, fcn);
      res = fcn(root->data[i - 1], ap);
      va_end(ap);
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}
/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */


/* backslash-region-begin */
#define _GCT_CONT_ARRAYS_FOREACH_FROM(name, algo, root, idx, ...)
  _GCT_CONT_LABEL_DECL(_gct_break);
  _gct_index_t     _gct_i;

  _GCT_ASSERT((_gct_sindex_t)idx <= (_gct_sindex_t)(root)->count);
  _GCT_ASSERT((_gct_sindex_t)idx >= -1);

  for (_gct_i = (idx + 1);
       _gct_i < (root)->count;
       _gct_i++)
    {
      _GCT_CONT_LABEL_DECL(_gct_continue);
      _GCT_UNUSED name##_index_t index = _gct_i;
      _GCT_UNUSED name##_item_t item = (root)->data[_gct_i];

      { __VA_ARGS__ }
      _GCT_CONT_LABEL(_gct_continue);
    }
  _GCT_CONT_LABEL(_gct_break);
/* backslash-region-end */



/* backslash-region-begin */
#define _GCT_CONT_ARRAYS_FOREACH_REVERSE_FROM(name, algo, root, idx, ...)
  _GCT_CONT_LABEL_DECL(_gct_break);
  _gct_index_t     _gct_i;

  _GCT_ASSERT((_gct_sindex_t)idx <= (_gct_sindex_t)(root)->count);
  _GCT_ASSERT((_gct_sindex_t)idx >= -1);

  for (_gct_i = (idx);
       _gct_i > 0;
       _gct_i--)
    {
      _GCT_CONT_LABEL_DECL(_gct_continue);
      _GCT_UNUSED name##_index_t index = _gct_i - 1;
      _GCT_UNUSED name##_item_t item = (root)->data[_gct_i - 1];

      { __VA_ARGS__ }
      _GCT_CONT_LABEL(_gct_continue);
    }
  _GCT_CONT_LABEL(_gct_break);
/* backslash-region-end */


#endif

