/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Binary trees common function and macro templates
*/

#ifndef _GCT_CONT_BINTREES_H_
#define _GCT_CONT_BINTREES_H_

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_SYMMETRIC_OPS(n, code)
  {
    { const uintptr_t n = 0; { code } }
    { const uintptr_t n = 1; { code } }
  }
/* backslash-region-end */

/***********************************************************************
 *      root access function
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_ROOT_FCNS(attr, name, algo, prefix, lockname)

_GCT_NONNULL(1)
attr GCT_CONT_ISORPHAN_PROTOTYPE(name, prefix)
{
  return !_GCT_CONT_ORPHAN_CHK(algo, name, _GCT_CONT_GET_ENTRY(name, item));
}

_GCT_NONNULL(1)
attr GCT_CONT_ORPHAN_PROTOTYPE(name, prefix)
{
  _GCT_CONT_ORPHAN_SET(algo, name, _GCT_CONT_GET_ENTRY(name, item));
}

attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)
{
  return index == NULL;
}

_GCT_NONNULL(1)
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)
{
  return root->root == NULL;
}

attr GCT_CONT_GET_PROTOTYPE(name, prefix)
{
  return index;
}

_GCT_NONNULL(1)
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)
{
  return root->count;
}

_GCT_NONNULL(1)
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)
{
  return -1;
}

_GCT_NONNULL(1)
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)
{
  root->root = NULL;
  root->count = 0;

  return _GCT_LOCK_INIT(lockname, &root->lock);
}

_GCT_NONNULL(1)
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_DESTROY(lockname, &root->lock);
}
/* backslash-region-end */


/***********************************************************************
 *      balance independent, parent link independent item access function
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_HEADTAIL_CODE(name, algo, prefix, lockname, istail)
{
  name##_entry_t *e = root->root;

  if (e == NULL)
    return NULL;

  while (e->child[istail] != NULL)
    e = e->child[istail];

  return _GCT_CONT_GET_ITEM(name, e);
}
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_ACCESS_FCNS(attr, name, algo, prefix, lockname, keyfield)

_GCT_NONNULL(1)
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_HEADTAIL_CODE(name, BINTREE, prefix, lockname, 0)

_GCT_NONNULL(1)
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_HEADTAIL_CODE(name, BINTREE, prefix, lockname, 1)

_GCT_NONNULL(1)
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)
{
  name##_entry_t *entry;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  entry = _GCT_CONT_GET_ENTRY(name, item);

  if (root->root == NULL)
    root->root = entry;
  else
    {
      name##_entry_t    *node = root->root;
      uint_fast8_t      dir;

      while (1)
        {
          name##_item_t node_i = _GCT_CONT_GET_ITEM(name, node);
          intmax_t res = _GCT_CONT_COMPARE_FIELDS(name, keyfield, node_i, item);
          dir = res < 0;

          if (node->child[dir] == NULL)
            break;

          node = node->child[dir];
        }

      node->child[dir] = entry;
      prefix##_gct_set_parent(entry, node);
    }

  root->count++;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return 1;
}

_GCT_NONNULL(1)
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)
{
  return prefix##_push(root, item);
}
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_BINTREES_ACCESS_UP_FCNS(attr, name, algo, prefix, lockname)

_GCT_NONNULL(1)
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_NEXTPREV_UP_CODE(name, BINTREE, prefix, lockname, 1)

_GCT_NONNULL(1)
attr GCT_CONT_PREV_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_NEXTPREV_UP_CODE(name, BINTREE, prefix, lockname, 0)

/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_BINTREES_ACCESS_NOUP_FCNS(attr, name, algo, prefix, lockname)

_GCT_NONNULL(1)
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_NEXTPREV_UP_CODE(name, BINTREE, prefix, lockname, 1) /* FIXME */

_GCT_NONNULL(1)
attr GCT_CONT_PREV_PROTOTYPE(name, prefix)
_GCT_CONT_BINTREES_NEXTPREV_UP_CODE(name, BINTREE, prefix, lockname, 0) /* FIXME */

/* backslash-region-end */

#endif

