/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Binary trees w/o parent link common function and macro templates
*/

#ifndef _GCT_CONT_BINTREES_NOUP_H_
#define _GCT_CONT_BINTREES_NOUP_H_

/***********************************************************************
 *      foreach macros
 */

/* backslash-region-begin */  /* In order iterative binary tree traversal using stack */
#define _GCT_CONT_BINTREES_NOUP_FOREACH_BASE(algo, name, root_, order, ...) \
									\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  _GCT_CONT_LABEL_DECL(_gct_continue);					\
  name##_entry_t *_gct_stack[name##_gct_stacksize];			\
  name##_entry_t *_gct_node = (root_)->root;				\
  uintptr_t i = 0;							\
									\
  while (_gct_node != NULL)						\
    {									\
      while (_gct_node != NULL)						\
        {								\
          if (_gct_node->child[!order] != NULL)				\
            {								\
              _GCT_ASSERT(i < name##_gct_stacksize);			\
              _gct_stack[i++] = _gct_node->child[!order];		\
            }								\
          _GCT_ASSERT(i < name##_gct_stacksize);			\
          _gct_stack[i++] = _gct_node;					\
          _gct_node = _gct_node->child[order];				\
        }								\
									\
      do								\
        {								\
          _GCT_CONT_LABEL_DECL(_gct_continue);				\
									\
          _GCT_ASSERT(i > 0);						\
          _gct_node = _gct_stack[--i];					\
									\
          {								\
            name##_item_t item = _GCT_CONT_GET_ITEM(name, _gct_node);	\
            _GCT_UNUSED name##_index_t index = item;			\
									\
            { __VA_ARGS__ }						\
          }								\
          _GCT_CONT_LABEL(_gct_continue);				\
        }								\
      while (i && _gct_node->child[!order] == NULL);			\
									\
      if (!i)								\
        break;								\
									\
      _GCT_ASSERT(i > 0);						\
      _gct_node = _gct_stack[--i];					\
    }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */


/* backslash-region-begin */  /* preorder iterative binary tree traversal using stack */
#define _GCT_CONT_BINTREES_NOUP_FOREACH_UNORDERED_BASE(algo, name, root_, ...) \
									\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  _GCT_CONT_LABEL_DECL(_gct_continue);					\
									\
  if ((root_)->root != NULL)						\
    {									\
      name##_entry_t    *_gct_stack[name##_gct_stacksize];		\
      _gct_index_t      _gct_i = 1;					\
									\
      _gct_stack[0] = (root_)->root;					\
									\
      while (_gct_i)							\
        {								\
          name##_entry_t        *_gct_node = _gct_stack[--_gct_i];	\
									\
          do								\
            {								\
              name##_entry_t *_gct_next = _gct_node->child[0];		\
              name##_item_t item = _GCT_CONT_GET_ITEM(name, _gct_node);	\
              _GCT_UNUSED name##_index_t index = item;			\
									\
              /* only push right links */				\
              if (_gct_node->child[1] != NULL)				\
                {							\
                  _GCT_ASSERT(_gct_i < name##_gct_stacksize);		\
                  _gct_stack[_gct_i++] = _gct_node->child[1];		\
                }							\
									\
              { __VA_ARGS__ }						\
									\
              _GCT_CONT_LABEL(_gct_continue);				\
									\
              _gct_node = _gct_next;					\
            }								\
          while (_gct_node != NULL);					\
        }								\
    }									\
									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */

/***********************************************************************
 *      internal functions for binary trees
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_NOUP_INTERNAL_FCNS(attr, name, prefix, keyfield) \
									\
_GCT_INTERNAL name##_entry_t *						\
prefix##_gct_get_parent(name##_root_t *root,				\
                     name##_entry_t *node)				\
{									\
  name##_item_t node_i = _GCT_CONT_GET_ITEM(name, node);		\
  name##_entry_t *prev, *entry = root->root;				\
									\
  while (1)								\
    {									\
      _GCT_ASSERT(entry != NULL);					\
      {									\
        name##_item_t item = _GCT_CONT_GET_ITEM(name, entry);		\
        intmax_t res = _GCT_CONT_COMPARE_FIELDS(name, keyfield, item, node_i); \
									\
        if (!res)							\
          return prev;							\
									\
        prev = entry;							\
        entry = entry->child[res <= 0];					\
      }									\
    }									\
}									\
									\
_GCT_INTERNAL void							\
prefix##_gct_set_parent(name##_entry_t *node,				\
                         name##_entry_t *parent)			\
{									\
}									\
/* backslash-region-end */


/***********************************************************************
 *      root access function
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREES_ROOT_NOUP_FCNS(attr, name, algo,		\
                                          prefix, lockname)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)				\
{									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  {									\
    _GCT_CONT_BINTREES_NOUP_FOREACH_UNORDERED_BASE(algo, name, root, {	\
      _GCT_CONT_ORPHAN_SET(algo, name, _gct_node);			\
      _GCT_CONT_REFDROP(name, item);					\
    });									\
  }									\
									\
  root->root = NULL;							\
  root->count = 0;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
/* backslash-region-end */


/***********************************************************************
 *      item access function
 */

#endif

