/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_CONT_BITMAP_H_
#define _GCT_CONT_BITMAP_H_

#define _GCT_CONT_BITMAP_ALIGN(x, b)    ((((x) - 1) | ((b) - 1)) + 1)

#define _GCT_CONT_BITMAP_ALIGN_SIZE(size, type)                 \
 (_GCT_CONT_BITMAP_ALIGN(size, sizeof(type) * 8) / (sizeof(type) * 8))

#define _GCT_CONT_BITMAP_WORDSIZE(name) (sizeof(name##_entry_t) * 8)
#define _GCT_CONT_BITMAP_WSDIV(name, x) (((_gct_index_t)x) / _GCT_CONT_BITMAP_WORDSIZE(name))
#define _GCT_CONT_BITMAP_WSMOD(name, x) (((_gct_index_t)x) % _GCT_CONT_BITMAP_WORDSIZE(name))
#define _GCT_CONT_BITMAP_WSMUL(name, x) (((_gct_index_t)x) * _GCT_CONT_BITMAP_WORDSIZE(name))

/* backslash-region-begin */
#define _GCT_CONT_BITMAPS_FCNS(attr, name, prefix,
                               lockname, algo, ...)

attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)
{
  return index == name##_null_index;
}

attr GCT_CONT_ISVALID_PROTOTYPE(name, prefix)
{
  return index >= 0 && index < _GCT_CONT_##algo##_BITCOUNT(root, name);
}

_GCT_NONNULL(1)
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)
{
  name##_index_t        i;
  gct_bool_t           res = 1;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = 0; res && i < _GCT_CONT_##algo##_WORDCOUNT(root, name); i++)
    if (root->data[i])
      {
	res = 0;
	break;
      }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_GET_PROTOTYPE(name, prefix)
{
  gct_bool_t           res;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  res = _GCT_BIT_TEST(name##_entry_t,
                      root->data[_GCT_CONT_BITMAP_WSDIV(name, index)],
                      _GCT_CONT_BITMAP_WSMOD(name, index));

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_SET_PROTOTYPE(name, prefix)
{
  gct_bool_t           res;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  res = _GCT_BIT_TEST(name##_entry_t,
                      root->data[_GCT_CONT_BITMAP_WSDIV(name, index)],
                      _GCT_CONT_BITMAP_WSMOD(name, index));

  _GCT_BIT_SETVALUE(name##_entry_t,
                    root->data[_GCT_CONT_BITMAP_WSDIV(name, index)],
                    _GCT_CONT_BITMAP_WSMOD(name, index), item);

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)
{
  return index < _GCT_CONT_##algo##_BITCOUNT(root, name)
    ? index + 1 : name##_null_index;
}

attr GCT_CONT_PREV_PROTOTYPE(name, prefix)
{
  return index > 0
    ? index - 1 : name##_null_index;
}

_GCT_NONNULL(1)
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix)
{
  return 0;
}

_GCT_NONNULL(1)
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix)
{
  return _GCT_CONT_##algo##_BITCOUNT(root, name) - 1;
}

_GCT_NONNULL(1)
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)
{
  name##_index_t        i;
  size_t                res = 0;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = 0; i < _GCT_CONT_##algo##_WORDCOUNT(root, name); i++)
    res += _GCT_POPCOUNT(root->data[i]);

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)
{
  return _GCT_CONT_##algo##_BITCOUNT(root, name);
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)
{
  _gct_index_t     i, j;
  _gct_sindex_t      res = 0;
  va_list       ap;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = 0; !res && i < _GCT_CONT_##algo##_WORDCOUNT(root, name); i++)
    {
      name##_entry_t    word;

      for (word = root->data[i];
           (j = _GCT_BIT_FFS(word));
           _GCT_BIT_CLR(name##_entry_t, word, j - 1))
        {
          va_start(ap, fcn);
          res = fcn(_GCT_CONT_BITMAP_WSMUL(name, i) + j - 1, ap);
          va_end(ap);
        }
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  for (i = 0; i < _GCT_CONT_##algo##_WORDCOUNT(root, name); i++)
    root->data[i] = 0;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);
}

/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_KEY_BITMAPS_FCNS(attr, name, prefix,
                                  lockname, algo, ...)

_GCT_NONNULL(1)
attr GCT_CONT_LOOKUP_PROTOTYPE(name, prefix, keyfield)
{
  _gct_index_t     i, j;
  _gct_sindex_t      res = name##_null_index;;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (i = 0; i < _GCT_CONT_##algo##_WORDCOUNT(root, name); i++)
    {
      name##_entry_t word = root->data[i];

      if (!value)
        word = ~word;

      if ((j = _GCT_BIT_FFS(word)))
        {
          res = _GCT_CONT_BITMAP_WSMUL(name, i) + j - 1;
          break;
        }
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1)
attr GCT_CONT_LOOKUP_NEXT_PROTOTYPE(name, prefix, keyfield)
{
  _gct_sindex_t res = name##_null_index;;
  name##_entry_t word;
  _gct_index_t i, j;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  if (++index < _GCT_CONT_##algo##_BITCOUNT(root, name))
    {
      name##_entry_t mask = ((name##_entry_t)1 << (_GCT_CONT_BITMAP_WSMOD(name, index))) - 1;
      i = _GCT_CONT_BITMAP_WSDIV(name, index);
      word = value ? root->data[i] & ~mask : root->data[i] | mask;

      while (1)
        {
          if (!value)
            word = ~word;

          if ((j = _GCT_BIT_FFS(word)))
            {
              res = _GCT_CONT_BITMAP_WSMUL(name, i) + j - 1;
              break;
            }

          if (++i == _GCT_CONT_##algo##_WORDCOUNT(root, name))
            break;

          word = root->data[i];
        }
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

/* backslash-region-end */




/* backslash-region-begin */
#define _GCT_CONT_BITMAPS_FOREACH(name, root, algo, ...)
  _GCT_CONT_LABEL_DECL(_gct_break);
  _gct_index_t     _gct_i;
  _gct_index_t     _gct_j;

  for (_gct_i = 0;
       _gct_i < _GCT_CONT_##algo##_WORDCOUNT(root, name);
       _gct_i++)
    {
      name##_entry_t    _gct_word;

      for (_gct_word = (root)->data[_gct_i];
           (_gct_j = _GCT_BIT_FFS(_gct_word));
           _GCT_BIT_CLR(name##_entry_t, _gct_word, _gct_j - 1))
        {
          _GCT_CONT_LABEL_DECL(_gct_continue);
          name##_index_t index = _GCT_CONT_BITMAP_WSMUL(name, _gct_i) + _gct_j - 1;
          _GCT_DEPRECATED _GCT_UNUSED name##_index_t item = index;

          { __VA_ARGS__ }
          _GCT_CONT_LABEL(_gct_continue);
        }
    }
  _GCT_CONT_LABEL(_gct_break);
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_BITMAPS_FOREACH_REVERSE(name, root, algo, ...)
  _GCT_CONT_LABEL_DECL(_gct_break);
  _gct_index_t     _gct_j;
  _gct_sindex_t    _gct_i;

  for (_gct_i = _GCT_CONT_##algo##_WORDCOUNT(root, name) - 1;
       _gct_i >= 0;
       _gct_i--)
    {
      name##_entry_t    _gct_word;

      for (_gct_word = (root)->data[_gct_i];
           (_gct_j = _GCT_BIT_FFSR(_gct_word));
           _GCT_BIT_CLR(name##_entry_t, _gct_word, _gct_j - 1))
        {
          _GCT_CONT_LABEL_DECL(_gct_continue);
          name##_index_t index = _GCT_CONT_BITMAP_WSMUL(name, _gct_i) + _gct_j - 1;
          _GCT_DEPRECATED _GCT_UNUSED name##_index_t item = index;

          { __VA_ARGS__ }
          _GCT_CONT_LABEL(_gct_continue);
        }
    }
  _GCT_CONT_LABEL(_gct_break);
/* backslash-region-end */


#endif


