/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    This file defines common helper macros for pointer based containers.

*/

#ifndef _GCT_CONT_ENTRY_H_
#define _GCT_CONT_ENTRY_H_

#include <gct/_platform.h>

/* "item from entry" and "entry from item" offset macros for pointer
   based containers */

#define _GCT_CONT_GET_ITEM(name, entry)                         \
 ((name##_item_t)(((uint8_t*)entry) - name##_gct_offsetof))

#define _GCT_CONT_GET_NULLITEM(name, entry)                     \
 (entry != NULL ? (name##_item_t)(((uint8_t*)entry) - name##_gct_offsetof) : NULL)

#define _GCT_CONT_GET_ENTRY(name, item)                         \
 ((name##_entry_t*)(((uint8_t*)item) + name##_gct_offsetof))

/* item orphan status shortcuts */

#define _GCT_CONT_ORPHAN_SET_(algo, name, entry) _GCT_CPP_CONCAT(_gct_orphan_##algo##_, name, _set)(entry)
#define _GCT_CONT_ORPHAN_SET(algo, name, entry) _GCT_CONT_ORPHAN_SET_(algo, GCT_CONTAINER_ORPHAN_CHK_##name, entry)

#define _GCT_CONT_ORPHAN_CHK_(algo, name, entry) _GCT_CPP_CONCAT(_gct_orphan_##algo##_, name, _chk)(entry)
#define _GCT_CONT_ORPHAN_CHK(algo, name, entry) _GCT_CONT_ORPHAN_CHK_(algo, GCT_CONTAINER_ORPHAN_CHK_##name, entry)

#define _GCT_CONT_ORPHAN_EMPTY_FCNS(algo, name, prefix)                               \
_GCT_INTERNAL void _gct_orphan_##algo##_##prefix##_set(name##_entry_t *e) {}          \
_GCT_INTERNAL gct_bool_t _gct_orphan_##algo##_##prefix##_chk(name##_entry_t *e) { return 1; }

/* container counter shortcuts */

#define _GCT_CONT_COUNTER_T_(name, prefix) _GCT_CPP_CONCAT(_gct_counter_##name##_, prefix, _t)
#define _GCT_CONT_COUNTER_T(name) _GCT_CONT_COUNTER_T_(name, GCT_CONTAINER_COUNTER_##name)

#define _GCT_CONT_COUNTER_ADD_(name, prefix, root, val) _GCT_CPP_CONCAT(_gct_counter_##name##_, prefix, _add)(root, val)
#define _GCT_CONT_COUNTER_ADD(name, root, val) _GCT_CONT_COUNTER_ADD_(name, GCT_CONTAINER_COUNTER_##name, root, val)

#define _GCT_CONT_COUNTER_SET_(name, prefix, root, val) _GCT_CPP_CONCAT(_gct_counter_##name##_, prefix, _set)(root, val)
#define _GCT_CONT_COUNTER_SET(name, root, val) _GCT_CONT_COUNTER_SET_(name, GCT_CONTAINER_COUNTER_##name, root, val)

#define _GCT_CONT_COUNTER_GET_(name, prefix, root) _GCT_CPP_CONCAT(_gct_counter_##name##_, prefix, _get)(root)
#define _GCT_CONT_COUNTER_GET(name, root) _GCT_CONT_COUNTER_GET_(name, GCT_CONTAINER_COUNTER_##name, root)

#define _GCT_CONT_COUNTER_ENABLED_(name, prefix) _GCT_CPP_CONCAT(_gct_counter_##name##_, prefix, _enabled)
#define _GCT_CONT_COUNTER_ENABLED(name) _GCT_CONT_COUNTER_ENABLED_(name, GCT_CONTAINER_COUNTER_##name)

/* backslash-region-begin */
#define _GCT_CONT_COUNTER_TYPES(name)

typedef _gct_empty_t _gct_counter_##name##_GCT_CONTAINER_COUNTER_##name##_t;
typedef size_t _gct_counter_##name##__t;

/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_COUNTER_FCNS(name)

static const gct_bool_t _gct_counter_##name##_GCT_CONTAINER_COUNTER_##name##_enabled = 0;
static const gct_bool_t _gct_counter_##name##__enabled = 1;

_GCT_INTERNAL void
_gct_counter_##name##_GCT_CONTAINER_COUNTER_##name##_add(name##_root_t *root, ssize_t diff)
{
}

_GCT_INTERNAL void
_gct_counter_##name##__add(name##_root_t *root, ssize_t diff)
{
	const void *foo = ((const void*)&root->counter);
  *((size_t*)foo) += diff;
}

_GCT_INTERNAL void
_gct_counter_##name##_GCT_CONTAINER_COUNTER_##name##_set(name##_root_t *root, size_t val)
{
}

_GCT_INTERNAL void
_gct_counter_##name##__set(name##_root_t *root, size_t val)
{
	const void *foo = ((const void*)&root->counter);
	*((size_t*)foo) = val;
}

_GCT_INTERNAL size_t
_gct_counter_##name##_GCT_CONTAINER_COUNTER_##name##_get(name##_root_t *root)
{
  return 0;
}

_GCT_INTERNAL size_t
_gct_counter_##name##__get(name##_root_t *root)
{
	const void *foo = ((const void*)&root->counter);
	return *((const size_t*)foo);
}
/* backslash-region-end */


#endif

