/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    This file defines container iteration macros

*/

#ifndef _GCT_CONT_FOREACH_H_
#define _GCT_CONT_FOREACH_H_

/* GCT_CONTAINER_FOREACH break and continue macros */

#if defined(__GNUC__) && !defined(__clang__) && !defined(__ICC)

#define _GCT_CONT_LABEL_DECL(x)         __label__ x
#define _GCT_CONT_LABEL(x)              goto x; x:
#define _GCT_CONT_GOTO(x)               goto x
#define GCT_FOREACH_BREAK         goto _gct_break
#define GCT_FOREACH_CONTINUE      goto _gct_continue

#else

#define _GCT_CONT_LABEL_DECL(x)
#define _GCT_CONT_LABEL(x)
#define _GCT_CONT_GOTO(x)
#define GCT_FOREACH_BREAK         _GCT_STR_ERROR("GCT_FOREACH_BREAK depends on missing GNU C extension");
#define GCT_FOREACH_CONTINUE      _GCT_STR_ERROR("GCT_FOREACH_CONTINUE depends on missing GNU C extension");

#endif

/* Iterate over the whole container and execute usercode for each
   item */

/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_NOLOCK(name, algo, root, ...)		\
{									\
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FOREACH)				\
                 (name, root, __VA_ARGS__)				\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_UNORDERED_NOLOCK(name, algo, root, ...)	\
{									\
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FOREACH_UNORDERED)			\
                 (name, root, __VA_ARGS__)				\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_REVERSE_NOLOCK(name, algo, root, ...)	\
{									\
  _GCT_CPP_CONCAT(_GCT_CONT_, algo, _FOREACH_REVERSE)			\
                 (name, root, __VA_ARGS__)				\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH(name, algo, root, ...)			\
{									\
  _GCT_LOCK_RDLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_NOLOCK(name, algo, root, __VA_ARGS__)		\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_UNORDERED(name, algo, root, ...)		\
{									\
  _GCT_LOCK_RDLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_UNORDERED_NOLOCK(name, algo, root, __VA_ARGS__)	\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_REVERSE(name, algo, root, ...)		\
{									\
  _GCT_LOCK_RDLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_REVERSE_NOLOCK(name, algo, root, __VA_ARGS__)	\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_WRLOCK(name, algo, root, ...)		\
{									\
  _GCT_LOCK_WRLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_NOLOCK(name, algo, root, __VA_ARGS__)		\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_UNORDERED_WRLOCK(name, algo, root, ...)	\
{									\
  _GCT_LOCK_WRLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_UNORDERED_NOLOCK(name, algo, root, __VA_ARGS__)	\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */


/* backslash-region-begin */
#define GCT_CONTAINER_FOREACH_REVERSE_WRLOCK(name, algo, root, ...)	\
{									\
  _GCT_LOCK_WRLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
  GCT_CONTAINER_FOREACH_REVERSE_NOLOCK(name, algo, root, __VA_ARGS__)	\
  _GCT_LOCK_UNLOCK(GCT_CONTAINER_LOCK_##name, &(root)->lock);		\
}									\
/* backslash-region-end */

#endif

