/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Hash and compare mechanisms

*/

#ifndef _GCT_HASH_COMPARE_H_
#define _GCT_HASH_COMPARE_H_

#include <string.h>
#include <ctype.h>

#include <gct/_platform.h>

/***********************************************************************
 *      Test, hash and compare algorithms
 */

typedef uintptr_t _gct_key_hash_t;

/* backslash-region-begin */
#define _GCT_HASH_LIB_PROTO(attr)					\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_1(const uint8_t *blob, size_t size);			\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_2(const uint16_t *blob, size_t size);			\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_4(const uint32_t *blob, size_t size);			\
									\
attr _gct_key_hash_t							\
_gct_str_hash(const char *str);						\
									\
attr _gct_key_hash_t							\
_gct_strcase_hash(const char *str);					\
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_HASH_LIB_FCNS(attr)					\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_1(const uint8_t *blob, size_t size)			\
{									\
  _gct_key_hash_t   val = 0;						\
  size_t            i;							\
									\
  for (i = 0; i < size; i++)						\
    val = blob[i] ^ (val << 1);						\
									\
  return val;								\
}									\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_2(const uint16_t *blob, size_t size)			\
{									\
  _gct_key_hash_t   val = 0;						\
  size_t            i;							\
									\
  for (i = 0; i < size / 2; i++)					\
    val = blob[i] ^ (val << 1);						\
									\
  return val;								\
}									\
									\
attr _gct_key_hash_t							\
_gct_rawdata_hash_4(const uint32_t *blob, size_t size)			\
{									\
  _gct_key_hash_t   val = 0;						\
  size_t            i;							\
									\
  for (i = 0; i < size / 4; i++)					\
    val = blob[i] ^ (val << 1);						\
									\
  return val;								\
}									\
									\
attr _gct_key_hash_t							\
_gct_str_hash(const char *str)						\
{									\
  _gct_key_hash_t       val = 0;					\
									\
  while (*str)								\
    val = (_gct_key_hash_t)*str++ ^ (val << 3) ^ val;			\
									\
  return val;								\
}									\
									\
attr _gct_key_hash_t							\
_gct_strcase_hash(const char *str)					\
{									\
  _gct_key_hash_t       val = 0;					\
									\
  while (*str)								\
    val = (_gct_key_hash_t)tolower(*str++) ^ (val << 3) ^ val;		\
									\
  return val;								\
}									\
/* backslash-region-end */

#ifdef GCT_CONFIG_USE_LIB
 _GCT_HASH_LIB_PROTO()
#else
 _GCT_HASH_LIB_FCNS(_GCT_UNUSED static)
#endif

/* raw data dispatch function */
_GCT_ALWAYS_INLINE _GCT_INTERNAL _gct_key_hash_t
_gct_rawdata_hash(const void *blob, size_t size)
{
  if (!(size & 3)
#ifdef GCT_CONFIG_NONALIGNED_ACCESS
      && !(addr & 3)
#endif
      )
    return _gct_rawdata_hash_4(blob, size);

  if (!(size & 1)
#ifdef GCT_CONFIG_NONALIGNED_ACCESS
      && !(addr & 1)
#endif
      )
    return _gct_rawdata_hash_2(blob, size);

  return _gct_rawdata_hash_1(blob, size);
}


/***********************************************************************
 *      hash and compare shortcut macros used by container code
 */

#define _GCT_CONT_HASH_ARG(name, keyfield, arg)             \
  _gct_##name##_##keyfield##_hash(_gct_##name##_##keyfield##_access_arg(arg))

#define _GCT_CONT_HASH_FIELD(name, keyfield, item)              \
  _gct_##name##_##keyfield##_hash(_gct_##name##_##keyfield##_access_field(item))

#define _GCT_CONT_COMPARE_FIELD_ARG(name, keyfield, item, arg)    \
  _gct_##name##_##keyfield##_compare(_gct_##name##_##keyfield##_access_field(item), \
                                     _gct_##name##_##keyfield##_access_arg(arg))

#define _GCT_CONT_COMPARE_FIELDS(name, keyfield, item1, item2)  \
  _gct_##name##_##keyfield##_compare(_gct_##name##_##keyfield##_access_field(item1), \
                                     _gct_##name##_##keyfield##_access_field(item2))


/***********************************************************************
 *      define hash and compare functions prototypes
 */

/* compare key values */
#define GCT_CONT_COMPARE_PROTOTYPE(name, keyfield)                         \
intmax_t                                                                \
_gct_##name##_##keyfield##_compare(_gct_##name##_##keyfield##_type_t a, \
                                   _gct_##name##_##keyfield##_type_t b)

/* compute key value hash */
#define GCT_CONT_HASH_PROTOTYPE(name, keyfield)                         \
_gct_key_hash_t                                                         \
_gct_##name##_##keyfield##_hash(_gct_##name##_##keyfield##_type_t a)

/* get key value from item ptr */
#define GCT_CONT_ACCESS_KF_PROTOTYPE(name, keyfield)                    \
_gct_##name##_##keyfield##_type_t                                       \
_gct_##name##_##keyfield##_access_field(name##_item_t name##_item)

/* get key value from passed lookup argument */
#define GCT_CONT_ACCESS_ARG_PROTOTYPE(name, keyfield)                   \
_gct_##name##_##keyfield##_type_t                                       \
_gct_##name##_##keyfield##_access_arg(_gct_##name##_##keyfield##_arg_t a)


/***********************************************************************
 *      contained field value access
 */

/*
 * Container item is the value to consider for hash and compare, no
 * field actually exist here
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_DIRECT_KF_ACCESS(name, testalgo, ...)		\
									\
typedef name##_item_t _gct_##name##__key_t;				\
									\
_GCT_CPP_CONCAT(_GCT_CONT_KEY_, testalgo, _TYPES)(name, , __VA_ARGS__);	\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_KF_PROTOTYPE(name, )	\
  { return name##_item; }						\
/* backslash-region-end */


/*
 * Container item is a struct and one of its member is the value to
 * consider for hash and compare.
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_MEMBER_KF_ACCESS(name, testalgo, keyfield, ...)	\
									\
typedef typeof((*(name##_item_t*)0).keyfield) _gct_##name##_##keyfield##_key_t; \
									\
_GCT_CPP_CONCAT(_GCT_CONT_KEY_, testalgo, _TYPES)(name, keyfield, __VA_ARGS__); \
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_KF_PROTOTYPE(name, keyfield) \
  { return name##_item.keyfield; }					\
/* backslash-region-end */


/*
 * Container item is a pointer to a struct and one of its member is
 * the value to consider for hash and compare.
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_PTR_KF_ACCESS(name, testalgo, keyfield, ...)	\
									\
typedef typeof(((name##_item_t)0)->keyfield) _gct_##name##_##keyfield##_key_t; \
									\
_GCT_CPP_CONCAT(_GCT_CONT_KEY_, testalgo, _TYPES)(name, keyfield, __VA_ARGS__); \
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_KF_PROTOTYPE(name, keyfield) \
  { return name##_item->keyfield; }					\
/* backslash-region-end */


/*
 * User code is provided to get value to consider for hash and compare.
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_CUSTOM_KF_ACCESS(name, testalgo, custom_expr, keyfield, ...) \
									\
typedef typeof(custom_expr) _gct_##name##_##keyfield##_key_t;		\
									\
_GCT_CPP_CONCAT(_GCT_CONT_KEY_, testalgo, _TYPES)(name, keyfield, __VA_ARGS__); \
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_KF_PROTOTYPE(name, keyfield) \
  { return custom_expr; }						\
/* backslash-region-end */

/***********************************************************************
 *       lookup function argument access mechanisms
 */

/*
 * Lookup value passed to the container access function has the same
 * type as the value in the container.
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)			\
									\
 typedef _gct_##name##_##keyfield##_type_t				\
  _gct_##name##_##keyfield##_arg_t;					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_ARG_PROTOTYPE(name, keyfield) \
  { return a; }								\
/* backslash-region-end */


/*
 * Lookup value passed to the container access function is a pointer
 * to an actual value with the same type as the value in the container.
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_PTR_ARG_ACCESS(name, keyfield)			\
									\
 typedef _gct_##name##_##keyfield##_type_t *				\
  _gct_##name##_##keyfield##_arg_t;					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_ACCESS_ARG_PROTOTYPE(name, keyfield) \
  { return *a; }							\
/* backslash-region-end */

/***********************************************************************
 *      hash and compare mechanisms
 */

/* FIXME intmax_t is bad (too long) */

/*
 * Scalar types with automatic cast,
 * item structure must be defined
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_SCALAR_TYPES(name, keyfield, ...)			\
 typedef _gct_##name##_##keyfield##_key_t				\
  _gct_##name##_##keyfield##_type_t;					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return (intmax_t)a - (intmax_t)b; }					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return a; }								\
									\
_GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */


/*
 * user specified scalar type
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_SCALARTYPE_TYPES(name, keyfield, type)		\
 typedef type _gct_##name##_##keyfield##_type_t;			\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return (intmax_t)a - (intmax_t)b; }					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return a; }								\
									\
_GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */


/*
 * string
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_STRING_TYPES(name, keyfield, ...)			\
 typedef const char * _gct_##name##_##keyfield##_type_t;		\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return strcmp(a, b); }						\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return _gct_str_hash(a); }						\
									\
_GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */


/*
 * string ignoring case
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_CASESTRING_TYPES(name, keyfield, ...)		\
 typedef const char * _gct_##name##_##keyfield##_type_t;		\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return strcasecmp(a, b); }						\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return _gct_strcase_hash(a); }					\
									\
_GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */


/*
 * aggregate
 * item structure must be defined
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_AGGREGATE_TYPES(name, keyfield, ...)		\
 typedef _gct_##name##_##keyfield##_key_t				\
   _gct_##name##_##keyfield##_type_t;					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return memcmp(&a, &b, sizeof(_gct_##name##_##keyfield##_key_t)); }	\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return _gct_rawdata_hash(&a, sizeof(_gct_##name##_##keyfield##_key_t)); } \
									\
_GCT_CONT_KEY_PTR_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */


/*
 * blob
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_BLOB_TYPES(name, keyfield, size)			\
 typedef const void * _gct_##name##_##keyfield##_type_t;		\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_COMPARE_PROTOTYPE(name, keyfield) \
  { return memcmp(a, b, size); }					\
									\
_GCT_ALWAYS_INLINE _GCT_INTERNAL GCT_CONT_HASH_PROTOTYPE(name, keyfield) \
  { return _gct_rawdata_hash(a, size); }				\
									\
_GCT_CONT_KEY_DIRECT_ARG_ACCESS(name, keyfield)				\
/* backslash-region-end */

#endif

