/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Key field based container functions set

*/

#ifndef _GCT_KEYFIELD_H_
#define _GCT_KEYFIELD_H_

#include <gct/_platform.h>
#include <gct/_cont_hash_cmp.h>

/***********************************************************************
 *      define all lookup and sorting functions prototypes
 */

/* Search the first item with a matching key field value */
/* backslash-block */
#define GCT_CONT_LOOKUP_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED name##_index_t						\
prefix##_lookup         (name##_root_t *root,				\
                         _gct_##name##_##keyfield##_arg_t value)	\

/* Search the first item with a matching key field value */
/* backslash-block */
#define GCT_CONT_REMOVE_KEY_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED gct_err_t							\
prefix##_remove_key     (name##_root_t *root,				\
                         _gct_##name##_##keyfield##_arg_t value)	\

/* Search the last item with a matching key field value */
/* backslash-block */
#define GCT_CONT_LOOKUP_LAST_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED name##_index_t						\
prefix##_lookup_last    (name##_root_t *root,				\
                         _gct_##name##_##keyfield##_arg_t value)	\

/* Search the next item with a matching key field value */
/* backslash-block */
#define GCT_CONT_LOOKUP_NEXT_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED name##_index_t						\
prefix##_lookup_next    (name##_root_t *root,				\
                         name##_index_t index,				\
                         _gct_##name##_##keyfield##_arg_t value)	\

/* Search the previous item with a matching key field value */
/* backslash-block */
#define GCT_CONT_LOOKUP_PREV_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED name##_index_t						\
prefix##_lookup_prev    (name##_root_t *root,				\
                         name##_index_t index,				\
                         _gct_##name##_##keyfield##_arg_t value)	\

/* Insert an item in ascending order sorted list */
/* backslash-block */
#define GCT_CONT_INSERT_ASCEND_PROTOTYPE(name, prefix, keyfield)	\
_GCT_UNUSED void							\
prefix##_insert_ascend  (name##_root_t *root,				\
                         name##_item_t item)				\

/* Insert an item in descending order sorted list */
/* backslash-block */
#define GCT_CONT_INSERT_DESCEND_PROTOTYPE(name, prefix, keyfield)	\
_GCT_UNUSED void							\
prefix##_insert_descend (name##_root_t *root,				\
                         name##_item_t item)				\

/* Sort items in container in ascending order */
/* backslash-block */
#define GCT_CONT_SORT_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED void							\
prefix##_sort_ascend    (name##_root_t *root)				\

/* Sort items in container in descending order */
/* backslash-block */
#define GCT_CONT_SORT_DESCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED void							\
prefix##_sort_descend   (name##_root_t *root)				\

/* Merge to containers sorted in ascending order */
/* backslash-block */
#define GCT_CONT_MERGE_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_UNUSED void							\
prefix##_merge_ascend    (name##_root_t *dest, name##_root_t *src)	\

/* Merge to containers sorted in descending order */
/* backslash-block */
#define GCT_CONT_MERGE_DESCEND_PROTOTYPE(name, prefix, keyfield)	\
_GCT_UNUSED void							\
prefix##_merge_descend   (name##_root_t *dest, name##_root_t *src)	\

/* Test if the container is sorted in ascending order */
/* backslash-block */
#define GCT_CONT_SORTED_ASCEND_PROTOTYPE(name, prefix, keyfield)	\
_GCT_UNUSED gct_bool_t							\
prefix##_issorted_ascend   (name##_root_t *root)			\

/* Test if the container is sorted in descending order */
/* backslash-block */
#define GCT_CONT_SORTED_DESCEND_PROTOTYPE(name, prefix, keyfield)	\
_GCT_UNUSED gct_bool_t							\
prefix##_issorted_descend   (name##_root_t *root)			\

/* Sort items in container in ascending order with a stable algorithm */
/* backslash-block */
#define GCT_CONT_SORT_STABLE_ASCEND_PROTOTYPE(name, prefix, keyfield)	\
void									\
prefix##_sort_stable_ascend    (name##_root_t *root)			\

/* Sort items in container in descending order with a stable algorithm */
/* backslash-block */
#define GCT_CONT_SORT_STABLE_DESCEND_PROTOTYPE(name, prefix, keyfield)	\
void									\
prefix##_sort_stable_descend   (name##_root_t *root)			\

/***********************************************************************
 *      User level lookup functions prototypes macro
 */

/* backslash-region-begin */
#define GCT_CONTAINER_KEY_PROTOTYPES(name, attr, prefix, keyfield)	\
									\
_GCT_UNUSED attr GCT_CONT_LOOKUP_PROTOTYPE(name, prefix, keyfield);	\
_GCT_UNUSED attr GCT_CONT_LOOKUP_LAST_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_LOOKUP_NEXT_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_LOOKUP_PREV_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_INSERT_ASCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_INSERT_DESCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_MERGE_ASCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_MERGE_DESCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_SORT_ASCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_SORT_DESCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_SORT_STABLE_ASCEND_PROTOTYPE(name, prefix, keyfield); \
_GCT_UNUSED attr GCT_CONT_SORT_STABLE_DESCEND_PROTOTYPE(name, prefix, keyfield); \
/* backslash-region-end */

/***********************************************************************
 *      User level container types and functions defintions macro
 */

/* backslash-block */
#define         GCT_CONTAINER_KEY_TYPES(name, access, testalgo, ...)	\
  _GCT_CPP_CONCAT(_GCT_CONT_KEY_, access, _KF_ACCESS)(name, testalgo, __VA_ARGS__); \

/* backslash-block */
#define         GCT_CONTAINER_KEY_FCNS(name, algo, attr, prefix, ...)	\
  _GCT_CPP_CONCAT(_GCT_CONT_KEY_, algo, _FCNS)(attr, name, prefix, GCT_CONTAINER_LOCK_##name, __VA_ARGS__) \

/* backslash-block */
#define         GCT_CONTAINER_KEY_NOLOCK_FCNS(name, algo, attr, prefix, ...) \
  _GCT_CPP_CONCAT(_GCT_CONT_KEY_, algo, _FCNS)(attr, name, prefix, NOLOCK, __VA_ARGS__) \

/* backslash-block */
#define         GCT_CONTAINER_KEY_LOCK_FCNS(name, algo, attr, prefix, lockname, ...) \
  _GCT_CPP_CONCAT(_GCT_CONT_KEY_, algo, _FCNS)(attr, name, prefix, lockname, __VA_ARGS__) \

#endif

