/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Linked lists common function and macro templates
*/

#ifndef _GCT_CONT_LISTS_H_
#define _GCT_CONT_LISTS_H_

#define _GCT_CONT_LIST_ISEND(algo, root, entry)                         \
 _GCT_CPP_CONCAT(_GCT_CONT_, algo, _ISEND)(root, entry)

#define _GCT_CONT_LIST_EMPTY(algo, root)                                \
 _GCT_CPP_CONCAT(_GCT_CONT_, algo, _EMPTY)(root)




/*
 * singly linked list previous entry and tail entry seek functions
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_INTERNAL_FCNS(attr, name, prefix)		\
									\
_GCT_INTERNAL name##_entry_t *						\
prefix##_gct_prev(name##_root_t *root,					\
                   name##_entry_t *ref)					\
{									\
  name##_entry_t        *entry = root->head;				\
									\
  if (entry == ref)							\
    return NULL;							\
									\
  for (; _GCT_ASSERT(entry != NULL), entry->next != ref;		\
       entry = entry->next)						\
         ;								\
									\
  return entry;								\
}									\
									\
_GCT_INTERNAL name##_entry_t *						\
prefix##_gct_tail(name##_root_t *root)					\
{									\
  name##_entry_t        *tail;						\
									\
  for (tail = root->head;						\
       tail && tail->next;						\
       tail = tail->next)						\
        ;								\
									\
  return tail;								\
}									\
									\
_GCT_INTERNAL name##_entry_t **						\
prefix##_gct_pprev(name##_root_t *root,					\
                    name##_entry_t *ref)				\
{									\
  name##_entry_t        **entry;					\
									\
  for (entry = &root->head; *entry != ref; entry = &(*entry)->next)	\
    ;									\
									\
  return entry;								\
}									\
									\
_GCT_INTERNAL name##_entry_t **						\
prefix##_gct_ptail(name##_root_t *root)					\
{									\
  name##_entry_t        **entry;					\
									\
  for (entry = &root->head;						\
       (*entry)->next != NULL;						\
       entry = &(*entry)->next)						\
    ;									\
									\
  return entry;								\
}									\
/* backslash-region-end */





/*
 * linked lists init common functions
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_INIT_FCNS(attr, name, algo, prefix, lockname)	\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)				\
{									\
  _GCT_CONT_LIST_EMPTY(algo, root);					\
									\
  return _GCT_LOCK_INIT(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)				\
{									\
  _GCT_LOCK_DESTROY(lockname, &root->lock);				\
}									\
/* backslash-region-end */






/*
 * linked lists container root common functions
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_ROOT_FCNS(attr, name, algo,			\
                                  prefix, lockname, head)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISORPHAN_PROTOTYPE(name, prefix)				\
{									\
  return !_GCT_CONT_ORPHAN_CHK(algo, name, _GCT_CONT_GET_ENTRY(name, item)); \
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ORPHAN_PROTOTYPE(name, prefix)				\
{									\
  _GCT_CONT_ORPHAN_SET(algo, name, _GCT_CONT_GET_ENTRY(name, item));	\
}									\
									\
attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)				\
{									\
  return index == NULL;							\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)				\
{									\
  return _GCT_CONT_LIST_ISEND(algo, root, root->head);			\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISFULL_PROTOTYPE(name, prefix)				\
{									\
  return 0;								\
}									\
									\
attr GCT_CONT_GET_PROTOTYPE(name, prefix)				\
{									\
  return index;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETPTR_PROTOTYPE(name, prefix)				\
{									\
  return index;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETINDEX_PROTOTYPE(name, prefix)				\
{									\
  return ptr;								\
}									\
/* backslash-region-end */





/*
 * direct access next/head and prev/tail functions
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_NEXT_HEAD_FCNS(attr, name, algo,		\
                                       prefix, lockname,		\
                                       NEXT, HEAD, next, head)		\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_##NEXT##_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                item;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  {									\
    name##_entry_t *next = _GCT_CONT_GET_ENTRY(name, index)->next;	\
									\
    item = !_GCT_CONT_LIST_ISEND(algo, root, next)			\
         ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, next))	\
         : NULL;							\
  }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_##HEAD##_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                item;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  item = !_GCT_CONT_LIST_ISEND(algo, root, root->head)			\
       ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, root->head))	\
       : NULL;								\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
/* backslash-region-end */





/*
 * indirect prev/tail access function for simple linked lists
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_PREV_TAIL_FCNS(attr, name, algo,		\
                                       prefix, lockname)		\
									\
_GCT_DEPRECATED								\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_PREV_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t        item;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  {									\
    name##_entry_t *index_e = _GCT_CONT_GET_ENTRY(name, index);		\
									\
    index_e = prefix##_gct_prev(root, index_e);				\
									\
    item = !_GCT_CONT_LIST_ISEND(algo, root, index_e)			\
         ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, index_e))	\
         : NULL;							\
  }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_DEPRECATED								\
_GCT_NONNULL(1)								\
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t        *tail;						\
  name##_index_t        item;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  tail = prefix##_gct_tail(root);					\
									\
  item = !_GCT_CONT_LIST_ISEND(algo, root, tail)			\
       ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, tail))		\
       : NULL;								\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
/* backslash-region-end */





/*
 * linked list entry count function
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_COUNT_FCNS(attr, name, algo,			\
                                   prefix, lockname, head)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)				\
{									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  name##_entry_t        *entry = root->head;				\
  size_t                i;						\
									\
  for (i = 0; !_GCT_CONT_LIST_ISEND(algo, root, entry); i++)		\
    entry = entry->next;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return i;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)				\
{									\
  return -1;								\
}									\
/* backslash-region-end */





/*
 * linked list clear functions
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_CLEAR_FCNS(attr, name, algo,			\
                                   prefix, lockname, head)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t        *entry, *next;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  for (entry = root->head;						\
       !_GCT_CONT_LIST_ISEND(algo, root, entry);			\
       entry = next)							\
    {									\
      next = entry->next;						\
      _GCT_CONT_ORPHAN_SET(algo, name, entry);				\
      _GCT_CONT_REFDROP(name, _GCT_CONT_GET_ITEM(name, entry));		\
    }									\
									\
  _GCT_CONT_LIST_EMPTY(algo, root);					\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
/* backslash-region-end */





/*
 * forward/backward iteration template function code
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_FOREACH_FUNCCODE(attr, name, algo, prefix,	\
                                         lockname, head, next)		\
									\
{									\
  name##_entry_t        *entry;						\
  intptr_t              res = 0;					\
  va_list               ap;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (entry = root->head;						\
       !res && !_GCT_CONT_LIST_ISEND(algo, root, entry);		\
       entry = entry->next)						\
    {									\
      va_start(ap, fcn);						\
      res = fcn(_GCT_CONT_GET_ITEM(name, entry), ap);			\
      va_end(ap);							\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */





/*
 * backward iteration template function code for singly linked lists
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_FOREACH_REVERSE_FUNCCODE(attr, name, algo,	\
                                                 prefix, lockname)	\
									\
{									\
  name##_entry_t        *entry, *next;					\
  intptr_t              res = 0;					\
  va_list               ap;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (next = NULL;							\
       next != (root)->head;						\
       next = entry)							\
    {									\
      for (entry = (root)->head;					\
           entry->next != next;						\
           entry = entry->next)						\
        ;								\
      {									\
        va_start(ap, fcn);						\
        res = fcn(_GCT_CONT_GET_ITEM(name, entry), ap);			\
        va_end(ap);							\
      }									\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */





/*
 * forward/backward iteration template macro
 */


/* backslash-region-begin */
#define _GCT_CONT_LISTS_FOREACH_BASE(algo, name, root, head, next, ...)	\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  name##_entry_t        *_gct_entry;					\
  name##_entry_t        *_gct_next;					\
									\
  for (_gct_entry = head;						\
       !_GCT_CONT_LIST_ISEND(algo, root, _gct_entry);			\
       _gct_entry = _gct_next)						\
    {									\
      _GCT_CONT_LABEL_DECL(_gct_continue);				\
      name##_item_t item = _GCT_CONT_GET_ITEM(name, _gct_entry);	\
      _GCT_UNUSED name##_index_t index = item;				\
      _gct_next = _gct_entry->next;					\
									\
      { __VA_ARGS__ }							\
      _GCT_CONT_LABEL(_gct_continue);					\
    }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */





/*
 * backward iteration template macro for singly linked lists
 */


/* backslash-region-begin */
#define _GCT_CONT_LISTS_FOREACH_REVERSE_BASE(name, root, last, ...)	\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  name##_entry_t        *_gct_entry;					\
  name##_entry_t        *_gct_next;					\
									\
  for (_gct_next = last;						\
       _gct_next != (root)->head;					\
       _gct_next = _gct_entry)						\
    {									\
      for (_gct_entry = (root)->head;					\
           _gct_entry->next != _gct_next;				\
           _gct_entry = _gct_entry->next)				\
        ;								\
      {									\
        _GCT_CONT_LABEL_DECL(_gct_continue);				\
        name##_item_t item = _GCT_CONT_GET_ITEM(name, _gct_entry);	\
        _GCT_UNUSED name##_index_t index = item;			\
									\
        { __VA_ARGS__ }							\
        _GCT_CONT_LABEL(_gct_continue);					\
      }									\
    }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */





/*
 * efficient, stack based, linked list in-place merge sort (stable)
 */
/* FIXME may be optionnaly moved in a library */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_SORT_FCNS(algo, entry_type)			\
									\
_GCT_INTERNAL entry_type *						\
_gct_##algo##_merge(entry_type *a, entry_type *b,			\
                    gct_bool_t (*cmp)(entry_type *a, entry_type *b))	\
{									\
  entry_type *first, *last;						\
									\
  if (cmp(a, b))							\
    a = (last = first = a)->next;					\
  else									\
    b = (last = first = b)->next;					\
									\
  while (a != NULL && b != NULL)					\
    if (cmp(a, b))							\
      a = (last = last->next = a)->next;				\
    else								\
      b = (last = last->next = b)->next;				\
									\
  last->next = a != NULL ? a : b;					\
  return first;								\
}									\
									\
_GCT_INTERNAL entry_type *						\
_gct_##algo##_sort(entry_type *tail,					\
                   gct_bool_t (*cmp)(entry_type *a, entry_type *b))	\
{									\
    size_t n = 0;							\
    entry_type *stack[32];      /* we are able to sort 2^31 nodes */	\
    entry_type **s = stack;						\
									\
    while (tail != NULL)						\
      {									\
        size_t idx, tmp;						\
        entry_type *a = tail;						\
        entry_type *b = tail->next;					\
									\
        if (b == NULL)							\
          {								\
            *s++ = a;							\
            break;							\
          }								\
									\
        tail = b->next;							\
									\
        if (cmp(a, b))							\
          ((*s = a)->next = b)->next = 0;				\
        else								\
          ((*s = b)->next = a)->next = 0;				\
        s++;								\
									\
        tmp = n++;							\
        for (idx = n ^ tmp; idx &= idx - 1; s--)			\
          s[-2] = _gct_##algo##_merge(s[-2], s[-1], cmp);		\
      }									\
									\
    while (s-- > stack + 1)						\
      s[-1] = _gct_##algo##_merge(s[-1], s[0], cmp);			\
									\
    return stack[0];							\
}									\
/* backslash-region-end */





/*
 * Comparaison functions for use with merge and sorting algorithm
 */

/* backslash-region-begin */
#define _GCT_CONT_LISTS_CMP_FCNS(name, prefix, algo, keyfield)		\
									\
_GCT_INTERNAL gct_bool_t						\
prefix##_gct_cmp_ascend(_gct_##algo##_entry_t *a,			\
                       _gct_##algo##_entry_t *b)			\
{									\
  return (_GCT_CONT_COMPARE_FIELDS(name, keyfield,			\
           _GCT_CONT_GET_ITEM(name, a),					\
           _GCT_CONT_GET_ITEM(name, b)) <= 0);				\
}									\
									\
_GCT_INTERNAL gct_bool_t						\
prefix##_gct_cmp_descend(_gct_##algo##_entry_t *a,			\
                        _gct_##algo##_entry_t *b)			\
{									\
  return (_GCT_CONT_COMPARE_FIELDS(name, keyfield,			\
           _GCT_CONT_GET_ITEM(name, a),					\
           _GCT_CONT_GET_ITEM(name, b)) >= 0);				\
}									\
/* backslash-region-end */





/* backslash-region-begin */
#define _GCT_CONT_LISTS_SORTED_FUNCCODE(algo, name, lockname, cmp, head) \
{									\
  name##_entry_t        *entry;						\
  gct_bool_t            res = 1;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (entry = head;							\
       !_GCT_CONT_LIST_ISEND(algo, root, entry) &&			\
       !_GCT_CONT_LIST_ISEND(algo, root, entry->next);			\
       entry = entry->next)						\
    {									\
      if (!cmp(entry, entry->next))					\
        {								\
          res = 0;							\
          break;							\
        }								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */


#endif

