/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Container locking stuff

*/

#ifndef _GCT_CONT_LOCK_H_
#define _GCT_CONT_LOCK_H_

#include <gct/_platform.h>

/* Lock the container root for modification */
#define GCT_CONT_WRLOCK_PROTOTYPE(name, prefix)                    \
_GCT_UNUSED void                                                            \
prefix##_wrlock         (name##_root_t *root)

/* Lock the container root for read */
#define GCT_CONT_RDLOCK_PROTOTYPE(name, prefix)                    \
_GCT_UNUSED void                                                            \
prefix##_rdlock         (name##_root_t *root)

/* Unlock the container root for read */
#define GCT_CONT_UNLOCK_PROTOTYPE(name, prefix)                    \
_GCT_UNUSED void                                                            \
prefix##_unlock         (name##_root_t *root)

/***********************************************************************/

#define _GCT_LOCK_TYPES(lockname)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _type_t)

#define _GCT_LOCK_INITIALIZER(lockname)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _initializer)

#define _GCT_LOCK_INIT(lockname, lock)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _init)(lock)

#define _GCT_LOCK_DESTROY(lockname, lock)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _destroy)(lock)

#define _GCT_LOCK_WRLOCK(lockname, lock)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _wrlock)(lock)

#define _GCT_LOCK_RDLOCK(lockname, lock)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _rdlock)(lock)

#define _GCT_LOCK_UNLOCK(lockname, lock)  _GCT_CPP_CONCAT(_gct_lock_, lockname, _unlock)(lock)

/***********************************************************************/

/* backslash-region-begin */
#define _GCT_CONT_LOCK_FCNS(attr, name, prefix, lockname)

attr GCT_CONT_WRLOCK_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_WRLOCK(lockname, &root->lock);
}

attr GCT_CONT_RDLOCK_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_RDLOCK(lockname, &root->lock);
}

attr GCT_CONT_UNLOCK_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_UNLOCK(lockname, &root->lock);
}
/* backslash-region-end */

/***********************************************************************/

typedef _gct_empty_t _gct_lock_NOLOCK_type_t;

#define _GCT_LOCK_NOLOCK_INITIALIZER    {}

/* backslash-region-begin */
#define _GCT_LOCK_EMPTY_LOCK_FCNS(lockname)
_GCT_INTERNAL void lockname##_wrlock(void *lock) {}
_GCT_INTERNAL void lockname##_rdlock(void *lock) {}
_GCT_INTERNAL void lockname##_unlock(void *lock) {}
_GCT_INTERNAL gct_err_t lockname##_init(void *lock) { return 0; }
_GCT_INTERNAL void lockname##_destroy(void *lock) {}
enum { lockname##_initializer = 0 };
/* backslash-region-end */

//static const _gct_empty_t lockname##_initializer = {};
_GCT_LOCK_EMPTY_LOCK_FCNS(_gct_lock_NOLOCK)

#define         GCT_LOCK_FEATURE(lockname, test) \
  _GCT_CPP_CONCAT(_GCT_LOCK_, lockname, _##test)

#endif

