/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_CONT_RING_H_
#define _GCT_CONT_RING_H_

_GCT_INTERNAL gct_bool_t
_gct_cont_ring_size_ispow2(size_t size)
{
  return !(size & (size - 1)); /* check size is a power of 2 */
}

/* backslash-region-begin */
#define _GCT_CONT_RINGS_FCNS(attr, name, prefix, lockname, algo)	\
									\
attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)				\
{									\
  return index == name##_null_index;					\
}									\
									\
attr GCT_CONT_ISVALID_PROTOTYPE(name, prefix)				\
{									\
  return ((index >= 0) && (index < root->count));			\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)				\
{									\
  return root->count == 0;						\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GET_PROTOTYPE(name, prefix)				\
{									\
  name##_item_t         item;						\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  item = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + (_gct_index_t)index)]; \
									\
  _GCT_CONT_REFNEW(name, item);						\
									\
  return item;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETPTR_PROTOTYPE(name, prefix)				\
{									\
  return root->data + _GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + (_gct_index_t)index); \
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETINDEX_PROTOTYPE(name, prefix)				\
{									\
  _gct_sindex_t      i = (ptr - root->data);				\
  _gct_sindex_t      n = i - _GCT_CONT_##algo##_SIZE_MOD(name, root, root->first); \
									\
  _GCT_ASSERT(!(((char*)ptr - (char*)root->data) % sizeof(*ptr)));	\
  _GCT_ASSERT(i >= 0);							\
  _GCT_ASSERT(i < _GCT_CONT_##algo##_SIZE(name, root));			\
									\
  if (n < 0)								\
    n += _GCT_CONT_##algo##_SIZE(name, root);				\
									\
  _GCT_ASSERT(n >= 0 && n < root->count);				\
									\
  return n;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_SET_PROTOTYPE(name, prefix)				\
{									\
  _gct_index_t     i;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  i = _GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + (_gct_index_t)index); \
  _GCT_CONT_REFDROP(name, root->data[i]);				\
  root->data[i] = _GCT_CONT_REFNEW(name, item);				\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return 0;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                res;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  res = index < root->count - 1						\
    ? _GCT_CONT_##algo##_SIZE_MOD(name, root, index + 1)		\
    : name##_null_index;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_PREV_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                res;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  res = index > 0							\
    ? _GCT_CONT_##algo##_SIZE_MOD(name, root, index - 1)		\
    : name##_null_index;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                res;					\
									\
  res = root->count ? 0 : name##_null_index;				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_TAIL_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                res;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  res = root->count							\
    ? _GCT_CONT_##algo##_SIZE_MOD(name, root, root->count - 1)		\
    : name##_null_index;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)				\
{									\
  return root->count;							\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)				\
{									\
  return _GCT_CONT_##algo##_SIZE(name, root);				\
}									\
									\
_GCT_INTERNAL void							\
prefix##_xchg_  (name##_item_t *a, name##_item_t *b)			\
{									\
  name##_item_t tmp;							\
									\
  tmp = *a;								\
  *a = *b;								\
  *b = tmp;								\
}									\
									\
attr GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix)			\
{									\
  size_t        res = 0;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  if (_gct_cont_ring_##name##_extend(root, 1))				\
    {									\
      _gct_index_t         i;						\
									\
      _GCT_CONT_REFNEW(name, item);					\
        /* FIXME replace with 2 memcpy/memmove */			\
      for (i = index; i < root->count; i++)				\
        prefix##_xchg_(&item, root->data + _GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + i)); \
									\
      root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + i)] = item; \
      root->count++;							\
      res++;								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
attr GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix)			\
{									\
  size_t        res = 0;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  if (_gct_cont_ring_##name##_extend(root, 1))				\
    {									\
      _gct_index_t         i;						\
									\
      _GCT_CONT_REFNEW(name, item);					\
        /* FIXME replace with 2 memcpy/memmove */			\
      for (i = index + 1; i < root->count; i++)				\
        prefix##_xchg_(&item, root->data + _GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + i)); \
									\
      root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + i)] = item; \
      root->count++;							\
      res++;								\
     }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)				\
{									\
  _gct_index_t     i;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  for (i = root->first; i != (_gct_index_t)(root->first + root->count); i++) \
    _GCT_CONT_REFDROP(name, root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, i)]); \
									\
  root->first = 0;							\
  root->count = 0;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix)				\
{									\
  _gct_index_t     j;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  _GCT_ASSERT((index >= 0) && (index < root->count));			\
									\
  _GCT_CONT_REFDROP(name, root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + index)]); \
									\
        /* FIXME replace with 2 memcpy/memmove */			\
  for (j = root->first + index;						\
       j != (_gct_index_t)(root->first + root->count - 1); j++)		\
    root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, j)]		\
      = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, j + 1)];	\
									\
  root->count--;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return 0;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)				\
{									\
  size_t        res = 0;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (_gct_cont_ring_##name##_extend(root, 1))				\
    {									\
      root->count++;							\
      root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, --root->first)] \
        = _GCT_CONT_REFNEW(name, item);					\
      res++;								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)				\
{									\
  size_t        res = 0;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (_gct_cont_ring_##name##_extend(root, 1))				\
    {									\
      root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + root->count++)] \
        = _GCT_CONT_REFNEW(name, item);					\
      res++;								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POP_PROTOTYPE(name, prefix)				\
{									\
  name##_item_t res = name##_null_item;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (root->count > 0)							\
    {									\
      root->count--;							\
      res = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first++)]; \
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix)				\
{									\
  name##_item_t res = name##_null_item;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (root->count > 0)							\
    res = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, root->first + --root->count)]; \
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_PUSH_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  _gct_index_t     i, j;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  i = _gct_cont_ring_##name##_extend(root, size);			\
									\
        /* FIXME replace with 2 memcpy/memmove */			\
  for (j = root->first;							\
       j != (_gct_index_t)(root->first - i); )				\
    root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, --j)]		\
                   = _GCT_CONT_REFNEW(name, *item++);			\
									\
  root->count += i;							\
  root->first -= i;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return i;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_PUSHBACK_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  _gct_index_t     i, j;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  i = _gct_cont_ring_##name##_extend(root, size);			\
									\
        /* FIXME replace with 2 memcpy/memmove */			\
  for (j = root->first + root->count;					\
       j != (_gct_index_t)(root->first + root->count + i); j++)		\
    root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, j)]		\
                 = _GCT_CONT_REFNEW(name, *item++);			\
									\
  root->count += i;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return i;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POP_ARRAY_PROTOTYPE(name, prefix)				\
{									\
  _gct_index_t     i, j;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  i = size > root->count ? root->count : size;				\
									\
        /* FIXME replace with 2 memcpy/memmove */			\
  for (j = root->first;							\
       j != (_gct_index_t)(root->first + i); j++)			\
    *item++ = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, j)];	\
									\
  root->first += i;							\
  root->count -= i;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return i;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POPBACK_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  _gct_index_t     i, j;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  i = size > root->count ? root->count : size;				\
									\
        /* FIXME replace with 2 memcpy/memmove */			\
  for (j = root->first + root->count;					\
       j != (_gct_index_t)(root->first + root->count - i); )		\
    *item++ = root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, --j)];	\
									\
  root->count -= i;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return i;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)				\
{									\
  _gct_index_t     i;							\
  intptr_t      res = 0;						\
  va_list       ap;							\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = root->first; !res &&						\
       i != (_gct_index_t)(root->first + root->count); i++)		\
    {									\
      va_start(ap, fcn);						\
      res = fcn(root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, i)], ap); \
      va_end(ap);							\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)			\
{									\
  _gct_index_t     i;							\
  intptr_t      res = 0;						\
  va_list       ap;							\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = root->first + root->count; !res && i != root->first; i--)	\
    {									\
      va_start(ap, fcn);						\
      res = fcn(root->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, i - 1)], ap); \
      va_end(ap);							\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */


/* backslash-region-begin */
#define _GCT_CONT_RINGS_FOREACH_FROM(name, algo, root, idx, ...)	\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  _gct_index_t     _gct_i;						\
									\
  _GCT_ASSERT((_gct_sindex_t)idx <= (_gct_sindex_t)(root)->count);	\
  _GCT_ASSERT((_gct_sindex_t)idx >= -1);				\
									\
  for (_gct_i = (root)->first + (idx + 1);				\
       _gct_i != (_gct_index_t)((root)->first + (root)->count);		\
       _gct_i++)							\
    {									\
      _GCT_CONT_LABEL_DECL(_gct_continue);				\
      _GCT_UNUSED name##_index_t index = _gct_i - (root)->first;	\
      _GCT_UNUSED name##_item_t item = (root)->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, _gct_i)]; \
									\
      { __VA_ARGS__ }							\
      _GCT_CONT_LABEL(_gct_continue);					\
    }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */



/* backslash-region-begin */
#define _GCT_CONT_RINGS_FOREACH_REVERSE_FROM(name, algo, root, idx, ...) \
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  _gct_index_t     _gct_i;						\
									\
  _GCT_ASSERT((_gct_sindex_t)idx <= (_gct_sindex_t)(root)->count);	\
  _GCT_ASSERT((_gct_sindex_t)idx >= -1);				\
									\
  for (_gct_i = (root)->first + (idx);					\
       _gct_i != (root)->first;						\
       _gct_i--)							\
    {									\
      _GCT_CONT_LABEL_DECL(_gct_continue);				\
      _GCT_UNUSED name##_index_t index = _gct_i - 1 - (root)->first;	\
      _GCT_UNUSED name##_item_t item = (root)->data[_GCT_CONT_##algo##_SIZE_MOD(name, root, _gct_i - 1)]; \
									\
      { __VA_ARGS__ }							\
      _GCT_CONT_LABEL(_gct_continue);					\
    }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */


#endif

