/*

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_PLATFORM_H_
#define _GCT_PLATFORM_H_

#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>

#define _GCT_STR_ERROR(str)     assert(((void)str, 0))

#define _GCT_ERROR_NA(name)     _GCT_STR_ERROR("operation not available for " #name)

#if defined(GCT_CONFIG_NOASSERT)
# define _GCT_ASSERT(v)
#else
# define _GCT_ASSERT(v)		assert(v)
#endif

#ifdef __GNUC__
# define _GCT_DEPRECATED        __attribute__ ((deprecated))
# define _GCT_ALWAYS_INLINE     __attribute__ ((always_inline))
# define _GCT_NONNULL(...)      __attribute__ ((nonnull(__VA_ARGS__)))
# define _GCT_UNUSED            __attribute__ ((unused))
#endif

#if (__STDC_VERSION__ >= 199901) || defined (__GNUC__)
# define _GCT_INLINE inline
#endif

/****/

#if defined(GCT_CONFIG_NODEPRECATED) || defined(GCT_CONFIG_NOWARNING)
# undef _GCT_DEPRECATED
#endif

/****/

#ifndef _GCT_DEPRECATED
# define _GCT_DEPRECATED
#endif

#ifndef _GCT_ALWAYS_INLINE
# define _GCT_ALWAYS_INLINE
#endif

#ifndef _GCT_NONNULL
# define _GCT_NONNULL(...)
#endif

#ifndef _GCT_UNUSED
# define _GCT_UNUSED
#endif

#ifndef _GCT_INLINE
# define _GCT_INLINE
#endif

#define _GCT_INTERNAL static _GCT_INLINE

/* use of intermediate macro allow macro expansion before concat */
#define _GCT_CPP_CONCAT(a, b, c)        a##b##c
#define _GCT_CPP_CONCAT2(a, b)          a##b

#if defined (__GNUC__) && !defined (__SUNPRO_C)
typedef struct {} _gct_empty_t;
#else
typedef char _gct_empty_t;
#endif

#ifndef GCT_CONFIG_NOPLATFORM

typedef int             gct_bool_t;
typedef int             gct_err_t;
typedef signed int      _gct_sindex_t; /* type for C array index */
typedef unsigned int    _gct_index_t; /* type for C array index */
#endif

#ifndef _GCT_UINT
#define _GCT_UINT unsigned int
#endif

#ifndef _GCT_INT
#define _GCT_INT signed int
#endif

#ifndef _GCT_ULONG
#define _GCT_ULONG unsigned long
#endif

#ifndef _GCT_ULONGLONG
#define _GCT_ULONGLONG unsigned long long
#endif

#endif

