/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_ARRAY_H_
#define _GCT_ARRAY_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_array.h>

#define _GCT_CONT_ARRAY_ARRAY_BASED      1
#define _GCT_CONT_ARRAY_POINTER_BASED    0
#define _GCT_CONT_ARRAY_ORDERED          1
#define _GCT_CONT_ARRAY_SORTED           0

/***********************************************************************
 *      types
 */

typedef _gct_empty_t _gct_ARRAY_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_ARRAY_TYPES(name, type, size)

typedef type                            name##_itembase_t;

typedef struct                          name##_root_s
{
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;
  size_t                                count;
  name##_itembase_t                     data[size];
}                                       name##_root_t;

typedef _gct_sindex_t                   name##_index_t;
typedef name##_itembase_t               name##_item_t;
typedef _gct_ARRAY_entry_t              name##_entry_t;

static const name##_item_t
name##_null_item = { 0 };

static const name##_index_t
name##_null_index = -1;

static const size_t
name##_gct_arraysize = size;

_GCT_INTERNAL size_t
_gct_cont_array_##name##_extend(name##_root_t *root, size_t s)
{
  size_t max = name##_gct_arraysize - root->count;
  return max > s ? s : max;
}
/* backslash-region-end */

#define _GCT_CONT_ARRAY_SIZE(name, root) name##_gct_arraysize

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_ARRAY_ROOT_INITIALIZER(name, lockname)
{
  .lock = _GCT_LOCK_INITIALIZER(lockname),
  .count = 0,
}
/* backslash-region-end */

#define _GCT_CONT_ARRAY_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_ARRAY_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

/* backslash-region-begin */
#define _GCT_CONT_ARRAY_FCNS(attr, name, prefix, lockname, ...)

_GCT_CONT_ARRAYS_FCNS(attr, name, prefix, lockname, ARRAY)

_GCT_NONNULL(1)
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)
{
  root->count = 0;

  return _GCT_LOCK_INIT(lockname, &root->lock);
}

_GCT_NONNULL(1)
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_DESTROY(lockname, &root->lock);
}

_GCT_NONNULL(1)
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)
{
  return !(root->count >= 0 && root->count <= name##_gct_arraysize);
}

_GCT_NONNULL(1)
attr GCT_CONT_ISFULL_PROTOTYPE(name, prefix)
{
  return root->count == name##_gct_arraysize;
}
/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */

#define _GCT_CONT_ARRAY_FOREACH_FROM(name, root, idx, ...)         \
 _GCT_CONT_ARRAYS_FOREACH_FROM(name, ARRAY, root, idx, __VA_ARGS__)

#define _GCT_CONT_ARRAY_FOREACH(name, root, ...)                   \
 _GCT_CONT_ARRAY_FOREACH_FROM(name, root, -1, __VA_ARGS__)

#define _GCT_CONT_ARRAY_FOREACH_REVERSE_FROM(name, root, idx, ...) \
 _GCT_CONT_ARRAYS_FOREACH_REVERSE_FROM(name, ARRAY, root, idx, __VA_ARGS__)

#define _GCT_CONT_ARRAY_FOREACH_REVERSE(name, root, ...)           \
 _GCT_CONT_ARRAY_FOREACH_REVERSE_FROM(name, root, (root)->count, __VA_ARGS__)

#define _GCT_CONT_ARRAY_FOREACH_UNORDERED(name, root, ...)         \
 _GCT_CONT_ARRAY_FOREACH(name, root, __VA_ARGS__)

/***********************************************************************
 *      key field based functions
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_ARRAY_FCNS(attr, name, prefix,
                                 lockname, keyfield, ...)
_GCT_CONT_LOOKUP_FCNS(attr, name, ARRAY, prefix,
                      lockname, keyfield);
/* backslash-region-end */


#endif

