/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_CONT_BINTREE_H_
#define _GCT_CONT_BINTREE_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_entry.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_pushpop_array.h>
#include <gct/_cont_bintree.h>
#include <gct/_cont_bintree_iter.h>
#include <gct/_cont_bintree_noup.h>

#define _GCT_CONT_BINTREE_ARRAY_BASED     0
#define _GCT_CONT_BINTREE_POINTER_BASED   1
#define _GCT_CONT_BINTREE_ORDERED         1
#define _GCT_CONT_BINTREE_SORTED          1

/***********************************************************************
 *      types
 */

typedef struct                          _gct_bintree_entry_s
{
  struct _gct_bintree_entry_s           *child[2];
}                                       _gct_BINTREE_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_BINTREE_TYPES(name, type, entryfield, stacksize)	\
									\
typedef struct                          name##_root_s			\
{									\
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;			\
  struct _gct_bintree_entry_s           *root;				\
  size_t                                count;				\
}                                       name##_root_t;			\
									\
typedef struct _gct_bintree_entry_s     name##_entry_t;			\
typedef type                            name##_itembase_t;		\
typedef name##_itembase_t *             name##_item_t;			\
typedef name##_itembase_t *             name##_index_t;			\
									\
static const name##_item_t						\
name##_null_item = NULL;						\
									\
static const name##_index_t						\
name##_null_index = NULL;						\
									\
static const uintptr_t							\
name##_gct_offsetof = offsetof(name##_itembase_t, entryfield);		\
									\
static const size_t							\
name##_gct_stacksize = stacksize;					\
									\
_GCT_CONT_ORPHAN_EMPTY_FCNS(BINTREE, name, GCT_CONTAINER_ORPHAN_CHK_##name) \
/* backslash-region-end */

/***********************************************************************
 *      orphan checking
 */

_GCT_INTERNAL void
_gct_orphan_BINTREE__set(struct _gct_bintree_entry_s *e)
{
  e->child[0] = e;
}

_GCT_INTERNAL gct_bool_t
_gct_orphan_BINTREE__chk(struct _gct_bintree_entry_s *e)
{
  return e->child[0] != e;
}

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREE_ROOT_INITIALIZER(name, lockname)		\
{									\
  .lock = _GCT_LOCK_INITIALIZER(lockname),				\
  .root = { .root = NULL }						\
}									\
/* backslash-region-end */

#define _GCT_CONT_BINTREE_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_BINTREE_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

/* backslash-region-begin */
#define _GCT_CONT_BINTREE_FCNS(attr, name, prefix, lockname, keyfield)	\
									\
_GCT_CONT_BINTREES_NOUP_INTERNAL_FCNS(attr, name, prefix, keyfield)	\
_GCT_CONT_BINTREES_ROOT_FCNS(attr, name, BINTREE, prefix, lockname)	\
_GCT_CONT_BINTREES_ACCESS_FCNS(attr, name, BINTREE, prefix, lockname, keyfield) \
_GCT_CONT_BINTREES_ROOT_NOUP_FCNS(attr, name, BINTREE, prefix, lockname) \
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)				\
{									\
  size_t count = 0;							\
  gct_err_t res;							\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  {									\
    _GCT_CONT_BINTREES_NOUP_FOREACH_UNORDERED_BASE(BINTREE, name, root, { \
      count++;								\
    });									\
  }									\
									\
  res = (count != root->count);						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
/* backslash-region-end */



/***********************************************************************
 *      iteration macros
 */

#define _GCT_CONT_BINTREE_FOREACH(name, root, ...)                 \
 _GCT_CONT_BINTREES_NOUP_FOREACH_BASE(BINTREE, name, root, 0, __VA_ARGS__)

#define _GCT_CONT_BINTREE_FOREACH_REVERSE(name, root, ...)         \
 _GCT_CONT_BINTREES_NOUP_FOREACH_BASE(BINTREE, name, root, 1, __VA_ARGS__)

#define _GCT_CONT_BINTREE_FOREACH_UNORDERED(name, root, ...)       \
 _GCT_CONT_BINTREES_NOUP_FOREACH_UNORDERED_BASE(BINTREE, name, root, __VA_ARGS__)

#endif

