/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_BITMAP_H_
#define _GCT_BITMAP_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_bitops.h>
#include <gct/_cont_bitmap.h>

#define _GCT_CONT_BITMAP_ARRAY_BASED      0
#define _GCT_CONT_BITMAP_POINTER_BASED    0
#define _GCT_CONT_BITMAP_ORDERED          1
#define _GCT_CONT_BITMAP_SORTED           0

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_TYPES(name, type, size, ...)

struct                                  name##_bitmap_s
{
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;
  type          data[_GCT_CONT_BITMAP_ALIGN_SIZE(size, type)];
};

typedef _gct_sindex_t                   name##_index_t;
typedef _gct_sindex_t                   name##_itembase_t;
typedef gct_bool_t                     name##_item_t;
typedef struct name##_bitmap_s          name##_root_t;
typedef type                            name##_entry_t;

static const size_t
name##_gct_size = size;

static const name##_index_t
name##_null_index = -1;

typedef gct_bool_t _gct_##name##_keyfield_arg_t;
/* backslash-region-end */

#define _GCT_CONT_BITMAP_BITCOUNT(root, name) (name##_gct_size)
#define _GCT_CONT_BITMAP_BYTECOUNT(root, name) (sizeof ((name##_root_t*)0)->data)
#define _GCT_CONT_BITMAP_WORDCOUNT(root, name) ((sizeof ((name##_root_t*)0)->data) / sizeof(name##_entry_t))

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_ROOT_INITIALIZER(name, lockname)
{
  .lock = _GCT_LOCK_INITIALIZER(lockname),
  .data = { 0 },
}
/* backslash-region-end */

#define _GCT_CONT_BITMAP_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_BITMAP_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_FCNS(attr, name, prefix, lockname, algo, ...)
 _GCT_CONT_BITMAPS_FCNS(attr, name, prefix, lockname, BITMAP)

_GCT_NONNULL(1)
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)
{
  _gct_index_t     i;

  for (i = 0; i < _GCT_CONT_BITMAP_WORDCOUNT(root, name); i++)
    root->data[i] = 0;

  return _GCT_LOCK_INIT(lockname, &root->lock);
}

_GCT_NONNULL(1)
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)
{
  _GCT_LOCK_DESTROY(lockname, &root->lock);
}

/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_KEY_BITMAP_FCNS(attr, name, prefix, lockname, algo)
 _GCT_CONT_KEY_BITMAPS_FCNS(attr, name, prefix, lockname, BITMAP)
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_FOREACH(name, root, ...)
 _GCT_CONT_BITMAPS_FOREACH(name, root, BITMAP, __VA_ARGS__)
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_FOREACH_REVERSE(name, root, ...)
 _GCT_CONT_BITMAPS_FOREACH_REVERSE(name, root, BITMAP, __VA_ARGS__)
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_BITMAP_FOREACH_UNORDERED(name, root, ...)
 _GCT_CONT_BITMAPS_FOREACH(name, root, BITMAP, __VA_ARGS__)
/* backslash-region-end */


#endif

