/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Circular Double Linked List template.

    The list is implemented with item structure as head/tail root.
*/

#ifndef _GCT_CONT_CLIST_H_
#define _GCT_CONT_CLIST_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_entry.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_pushpop_array.h>
#include <gct/_cont_lists.h>

#define _GCT_CONT_CLIST_ARRAY_BASED     0
#define _GCT_CONT_CLIST_POINTER_BASED   1
#define _GCT_CONT_CLIST_ORDERED         1
#define _GCT_CONT_CLIST_SORTED          0
#define _GCT_CONT_CLIST_REORDER         1

/***********************************************************************
 *      types
 */

typedef struct                          _gct_clist_entry_s
{
  struct _gct_clist_entry_s             *next;
  struct _gct_clist_entry_s             *prev;
}                                       _gct_CLIST_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_CLIST_TYPES(name, type, entryfield)			\
									\
typedef struct                          name##_root_s			\
{									\
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;			\
  struct _gct_clist_entry_s     ht;					\
}                                       name##_root_t;			\
									\
typedef struct _gct_clist_entry_s       name##_entry_t;			\
typedef type                            name##_itembase_t;		\
typedef name##_itembase_t *             name##_item_t;			\
typedef name##_itembase_t *             name##_index_t;			\
									\
static const name##_item_t						\
name##_null_item = NULL;						\
									\
static const name##_index_t						\
name##_null_index = NULL;						\
									\
static const uintptr_t							\
name##_gct_offsetof = offsetof(name##_itembase_t, entryfield);		\
									\
_GCT_CONT_ORPHAN_EMPTY_FCNS(CLIST, name, GCT_CONTAINER_ORPHAN_CHK_##name) \
/* backslash-region-end */


/***********************************************************************
 *      orphan checking
 */

_GCT_INTERNAL void
_gct_orphan_CLIST__set(struct _gct_clist_entry_s *e)
{
  e->next = NULL;
}

_GCT_INTERNAL gct_bool_t
_gct_orphan_CLIST__chk(struct _gct_clist_entry_s *e)
{
  return e->next != NULL;
}

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_ROOT_DECLARATOR(name, lockname, symbol)		\
name##_root_t symbol =							\
{									\
  .lock = _GCT_LOCK_INITIALIZER(lockname),				\
  .ht = { .next = &symbol.ht, .prev = &symbol.ht }			\
}									\
/* backslash-region-end */

/***********************************************************************
 *      access functions
 */

#define _GCT_CONT_CLIST_ISEND(root, entry)      ((entry) == &(root)->ht)
#define _GCT_CONT_CLIST_EMPTY(root)             ((root)->ht.next = (root)->ht.prev = &(root)->ht)

/* backslash-region-begin */
#define _GCT_CONT_CLIST_FCNS(attr, name, prefix, lockname, ...)		\
									\
_GCT_CONT_LISTS_INIT_FCNS(attr, name, CLIST, prefix, lockname)		\
									\
_GCT_CONT_LISTS_ROOT_FCNS(attr, name, CLIST, prefix, lockname, ht.next)	\
									\
_GCT_NONNULL(1, 2, 3)							\
attr GCT_CONT_SET_PROTOTYPE(name, prefix)				\
{									\
  gct_err_t          res = -1;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  {									\
    name##_entry_t *index_e = _GCT_CONT_GET_ENTRY(name, index);		\
    name##_entry_t *item_e = _GCT_CONT_GET_ENTRY(name, item);		\
									\
    if (_GCT_CONT_ORPHAN_CHK(CLIST, name, index_e))			\
      {									\
        _GCT_CONT_REFNEW(name, item);					\
									\
        (item_e->prev = index_e->prev)->next = item_e;			\
        (item_e->next = index_e->next)->prev = item_e;			\
									\
        _GCT_CONT_ORPHAN_SET(CLIST, name, index_e);			\
        _GCT_CONT_REFDROP(name, index);					\
        res = 0;							\
      }									\
  }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_CONT_LISTS_NEXT_HEAD_FCNS(attr, name, CLIST,			\
                               prefix, lockname,			\
                               NEXT, HEAD, next, ht.next)		\
									\
_GCT_CONT_LISTS_NEXT_HEAD_FCNS(attr, name, CLIST,			\
                               prefix, lockname,			\
                               PREV, TAIL, prev, ht.prev)		\
									\
_GCT_CONT_LISTS_COUNT_FCNS(attr, name, CLIST, prefix,			\
                           lockname, ht.next)				\
									\
_GCT_NONNULL(1, 2, 3)							\
attr GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix)			\
{									\
  name##_entry_t        *index_e;					\
  size_t                res = 0;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  index_e = _GCT_CONT_GET_ENTRY(name, index);				\
									\
  if (_GCT_CONT_ORPHAN_CHK(CLIST, name, index_e))			\
    {									\
      name##_entry_t    *item_e;					\
									\
      _GCT_CONT_REFNEW(name, item);					\
									\
      item_e = _GCT_CONT_GET_ENTRY(name, item);				\
      item_e->next = index_e;						\
      item_e->prev = index_e->prev;					\
      index_e->prev->next = item_e;					\
      index_e->prev = item_e;						\
									\
      res = 1;								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2, 3)							\
attr GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix)			\
{									\
  name##_entry_t        *index_e;					\
  size_t                res = 0;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  index_e = _GCT_CONT_GET_ENTRY(name, index);				\
									\
  if (_GCT_CONT_ORPHAN_CHK(CLIST, name, index_e))			\
    {									\
      name##_entry_t    *item_e;					\
									\
      _GCT_CONT_REFNEW(name, item);					\
									\
      item_e = _GCT_CONT_GET_ENTRY(name, item);				\
      item_e->prev = index_e;						\
      item_e->next = index_e->next;					\
      index_e->next->prev = item_e;					\
      index_e->next = item_e;						\
									\
      res = 1;								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_CONT_LISTS_CLEAR_FCNS(attr, name, CLIST,				\
                           prefix, lockname, ht.next)			\
									\
_GCT_INTERNAL void							\
prefix##_remove_        (name##_root_t *root,				\
                         name##_entry_t *entry)				\
{									\
  entry->prev->next = entry->next;					\
  entry->next->prev = entry->prev;					\
									\
  _GCT_CONT_ORPHAN_SET(CLIST, name, entry);				\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix)				\
{									\
  gct_err_t  res = -1;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  {									\
    name##_entry_t      *index_e = _GCT_CONT_GET_ENTRY(name, index);	\
									\
    if (_GCT_CONT_ORPHAN_CHK(CLIST, name, index_e))			\
      {									\
        prefix##_remove_(root, index_e);				\
        _GCT_CONT_REFDROP(name, index);					\
									\
        res = 0;							\
      }									\
									\
    _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
  }									\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t    *item_e;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
  _GCT_CONT_REFNEW(name, item);						\
									\
  item_e = _GCT_CONT_GET_ENTRY(name, item);				\
  item_e->prev = &root->ht;						\
  root->ht.next->prev = item_e;						\
  item_e->next = root->ht.next;						\
  root->ht.next = item_e;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return 1;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t    *item_e;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
  _GCT_CONT_REFNEW(name, item);						\
									\
  item_e = _GCT_CONT_GET_ENTRY(name, item);				\
  item_e->next = &root->ht;						\
  root->ht.prev->next = item_e;						\
  item_e->prev = root->ht.prev;						\
  root->ht.prev = item_e;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return 1;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POP_PROTOTYPE(name, prefix)				\
{									\
  name##_item_t         item = 0;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if ((root->ht.next != &root->ht))					\
    {									\
      name##_entry_t    *head = root->ht.next;				\
									\
      item = _GCT_CONT_GET_ITEM(name, head);				\
      prefix##_remove_(root, head);					\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix)				\
{									\
  name##_item_t         item = 0;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if ((root->ht.next != &root->ht))					\
    {									\
      name##_entry_t    *tail = root->ht.prev;				\
									\
      item = _GCT_CONT_GET_ITEM(name, tail);				\
      prefix##_remove_(root, tail);					\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_CONT_PUSHPOP_ARRAY_FCNS(attr, name, prefix, prefix, lockname);	\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)				\
_GCT_CONT_LISTS_FOREACH_FUNCCODE(attr, name, CLIST, prefix,		\
                                 lockname, ht.next, next)		\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)			\
_GCT_CONT_LISTS_FOREACH_FUNCCODE(attr, name, CLIST, prefix,		\
                                 lockname, ht.prev, prev)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t        *entry;						\
  gct_bool_t           res = 0;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  entry = &root->ht;							\
  do									\
    {									\
      if (entry->next == NULL)						\
        res |= 1;							\
									\
      if (entry->prev == NULL)						\
        res |= 2;							\
									\
      if (entry->next->prev != entry)					\
        res |= 4;							\
									\
      if (entry->prev->next != entry)					\
        res |= 8;							\
									\
      entry = entry->next;						\
    }									\
  while (!res && (entry != &root->ht));					\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */



/***********************************************************************
 *      iteration macros
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_FOREACH(name, root, ...)			\
 _GCT_CONT_LISTS_FOREACH_BASE(CLIST, name, root,			\
                              (root)->ht.next, next, __VA_ARGS__)	\
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_CLIST_FOREACH_FROM(name, root, index, ...)		\
 _GCT_CONT_LISTS_FOREACH_BASE(CLIST, name, root,			\
                              _GCT_CONT_GET_ENTRY(name, index)->next, next, __VA_ARGS__) \
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_CLIST_FOREACH_REVERSE(name, root, ...)		\
 _GCT_CONT_LISTS_FOREACH_BASE(CLIST, name, root,			\
                              (root)->ht.prev, prev, __VA_ARGS__)	\
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_CLIST_FOREACH_REVERSE_FROM(name, root, index, ...)	\
 _GCT_CONT_LISTS_FOREACH_BASE(CLIST, name, root,			\
                              _GCT_CONT_GET_ENTRY(name, index)->prev, prev, __VA_ARGS__) \
/* backslash-region-end */


#define _GCT_CONT_CLIST_FOREACH_UNORDERED(name, root, ...)         \
 _GCT_CONT_CLIST_FOREACH(name, root, __VA_ARGS__)


/***********************************************************************
 *      key field based functions
 */

/* linked list merge/sort functions */
_GCT_CONT_LISTS_SORT_FCNS(CLIST, _gct_CLIST_entry_t);




/*
 * CLIST links rebuild after merge/sort template code
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_REBUILD_CODE(name, lockname, cmp, entry, root)	\
{									\
  name##_entry_t *prev;							\
									\
  root->ht.next = entry;						\
									\
  /* rebuild prev links */						\
  for (prev = &root->ht ; entry != NULL; entry = entry->next)		\
    {									\
      entry->prev = prev;						\
      prev = entry;							\
    }									\
									\
  /* update links with root */						\
  prev->next = &root->ht;						\
  root->ht.prev = prev;							\
}									\
/* backslash-region-end */





/*
 * CLIST sort glue function template, call sort and rebuild prev links
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_SORT_FUNCCODE(name, lockname, cmp)		\
{									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (root->ht.next != root->ht.prev)					\
    {									\
      name##_entry_t *tmp;						\
									\
      /* terminate list with a NULL for sort */				\
      root->ht.prev->next = NULL;					\
      tmp = root->ht.next = _gct_CLIST_sort(root->ht.next, cmp);	\
									\
      _GCT_CONT_CLIST_REBUILD_CODE(name, lockname, cmp, tmp, root);	\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
/* backslash-region-end */





/*
 * CLIST merge glue function template, call merge and rebuild prev links
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_MERGE_FUNCCODE(name, lockname, cmp)		\
{									\
  _GCT_LOCK_WRLOCK(lockname, &src->lock);				\
  _GCT_LOCK_WRLOCK(lockname, &dest->lock);				\
									\
  if (src->ht.next != &src->ht)						\
    {									\
      if (dest->ht.next == &dest->ht)					\
        {								\
          /* move from src to dest */					\
          dest->ht = src->ht;						\
          dest->ht.prev->next =						\
            dest->ht.next->prev = &dest->ht;				\
        }								\
      else								\
        {								\
          name##_entry_t *tmp;						\
									\
          /* terminate lists with a NULL for merge operation */		\
          src->ht.prev->next = NULL;					\
          dest->ht.prev->next = NULL;					\
									\
          tmp = _gct_CLIST_merge(dest->ht.next, src->ht.next, cmp);	\
									\
          _GCT_CONT_CLIST_REBUILD_CODE(name, lockname, cmp, tmp, dest);	\
        }								\
									\
      _GCT_CONT_CLIST_EMPTY(src);					\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &dest->lock);				\
  _GCT_LOCK_UNLOCK(lockname, &src->lock);				\
}									\
/* backslash-region-end */





/*
 * CLIST Ordered insertion function template
 */

/* backslash-region-begin */
#define _GCT_CONT_CLIST_ORDERED_INSERT_FUNCCODE(name, lockname, cmp)	\
{									\
  name##_entry_t        *index_e, *item_e;				\
  size_t                res = 0;					\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  item_e = _GCT_CONT_GET_ENTRY(name, item);				\
									\
  for (index_e = root->ht.next; index_e != &root->ht; index_e = index_e->next) \
    if (cmp(item_e, index_e))						\
      break;								\
									\
  _GCT_CONT_REFNEW(name, item);						\
									\
  item_e->next = index_e;						\
  item_e->prev = index_e->prev;						\
  index_e->prev->next = item_e;						\
  index_e->prev = item_e;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
/* backslash-region-end */





/* backslash-region-begin */
#define _GCT_CONT_KEY_CLIST_FCNS(attr, name, prefix,			\
                                 lockname, keyfield, ...)		\
									\
_GCT_CONT_LOOKUP_FCNS(attr, name, CLIST, prefix, lockname, keyfield);	\
									\
_GCT_CONT_LISTS_CMP_FCNS(name, prefix, CLIST, keyfield)			\
									\
attr GCT_CONT_SORT_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_SORT_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend)	\
									\
attr GCT_CONT_SORT_DESCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_SORT_FUNCCODE(name, lockname, prefix##_gct_cmp_descend)	\
									\
attr GCT_CONT_SORT_STABLE_ASCEND_PROTOTYPE(name, prefix, keyfield)	\
{									\
  prefix##_sort_ascend(root);						\
}									\
									\
attr GCT_CONT_SORT_STABLE_DESCEND_PROTOTYPE(name, prefix, keyfield)	\
{									\
  prefix##_sort_descend(root);						\
}									\
									\
attr GCT_CONT_SORTED_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_LISTS_SORTED_FUNCCODE(CLIST, name, lockname, prefix##_gct_cmp_ascend, root->ht.next) \
									\
attr GCT_CONT_SORTED_DESCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_LISTS_SORTED_FUNCCODE(CLIST, name, lockname, prefix##_gct_cmp_descend, root->ht.next) \
									\
attr GCT_CONT_INSERT_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_ORDERED_INSERT_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend) \
									\
attr GCT_CONT_INSERT_DESCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_ORDERED_INSERT_FUNCCODE(name, lockname, prefix##_gct_cmp_descend) \
									\
attr GCT_CONT_MERGE_ASCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_MERGE_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend)	\
									\
attr GCT_CONT_MERGE_DESCEND_PROTOTYPE(name, prefix, keyfield)		\
_GCT_CONT_CLIST_MERGE_FUNCCODE(name, lockname, prefix##_gct_cmp_descend) \
/* backslash-region-end */



#endif

