/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_DARRAY_H_
#define _GCT_DARRAY_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_array.h>
#include <gct/_helpers.h>

#define _GCT_CONT_DARRAY_ARRAY_BASED      1
#define _GCT_CONT_DARRAY_POINTER_BASED    0
#define _GCT_CONT_DARRAY_ORDERED          1
#define _GCT_CONT_DARRAY_SORTED           0

/***********************************************************************
 *      types
 */

typedef _gct_empty_t _gct_DARRAY_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_DARRAY_TYPES(name, type, initialsize, maxsize)	\
									\
typedef type                            name##_itembase_t;		\
									\
typedef struct                          name##_root_s			\
{									\
  size_t                                count;				\
  size_t                                size;				\
  name##_itembase_t                     *data;				\
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;			\
}                                       name##_root_t;			\
									\
typedef _gct_sindex_t                   name##_index_t;			\
typedef name##_itembase_t               name##_item_t;			\
typedef _gct_DARRAY_entry_t              name##_entry_t;		\
									\
static const name##_item_t						\
name##_null_item = { 0 };						\
									\
static const name##_index_t						\
name##_null_index = -1;							\
									\
static const size_t							\
name##_gct_arrayinitialsize = initialsize;				\
									\
static const size_t							\
name##_gct_arraymaxsize = maxsize;					\
									\
_GCT_INTERNAL size_t							\
_gct_cont_array_##name##_extend(name##_root_t *root, size_t s)		\
{									\
  size_t res;								\
									\
  _GCT_ASSERT(name##_gct_arrayinitialsize > 0);				\
									\
  if (root->count + s <= root->size)					\
    return s;								\
									\
  if (root->size < maxsize)						\
    {									\
      size_t newsize = root->size ? root->size * 2 : initialsize;	\
      name##_itembase_t *data = _gct_realloc(root->data, newsize * sizeof (name##_item_t)); \
									\
      if (data != NULL)							\
        {								\
          root->data = data;						\
          root->size = newsize;						\
        }								\
    }									\
									\
  res = root->size - root->count;					\
  return res > s ? s : res;						\
}									\
/* backslash-region-end */

#define _GCT_CONT_DARRAY_SIZE(name, root) ((root)->size)

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_DARRAY_ROOT_INITIALIZER(name, lockname)		\
{									\
  .count = 0,								\
  .size = 0,								\
  .data = NULL,								\
  .lock = _GCT_LOCK_INITIALIZER(lockname),				\
}									\
/* backslash-region-end */

#define _GCT_CONT_DARRAY_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_DARRAY_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

/* backslash-region-begin */
#define _GCT_CONT_DARRAY_FCNS(attr, name, prefix, lockname, ...)	\
									\
_GCT_CONT_ARRAYS_FCNS(attr, name, prefix, lockname, DARRAY)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)				\
{									\
  if (name##_gct_arrayinitialsize > name##_gct_arraymaxsize)		\
    _GCT_STR_ERROR("DARRAY minimal size > maximal size");		\
									\
  root->count = 0;							\
  root->size = 0;							\
  root->data = NULL;							\
									\
  return _GCT_LOCK_INIT(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)				\
{									\
  if (root->data)							\
    {									\
      _gct_free(root->data);						\
      root->data = NULL;						\
    }									\
									\
  _GCT_LOCK_DESTROY(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_RESIZE_PROTOTYPE(name, prefix)				\
{									\
  name##_itembase_t *data;						\
  gct_err_t res = 0;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if (size > name##_gct_arraymaxsize)					\
    res = -1;								\
  else if (size < root->count)						\
    res = -1;								\
  else if (size == 0)							\
    {									\
      free(root->data);							\
      root->size = 0;							\
      root->data = NULL;						\
    }									\
  else									\
    {									\
      data = _gct_realloc(root->data, size * sizeof (name##_item_t));	\
									\
      if (data != NULL)							\
        {								\
          root->data = data;						\
          root->size = size;						\
        }								\
      else								\
        res = -1;							\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)				\
{									\
  if (root->size == 0)							\
    return 0;								\
									\
  return !(root->data != NULL						\
           && root->count >= 0						\
           && root->count <= root->size					\
           && root->size <= name##_gct_arraymaxsize			\
           && root->size >= name##_gct_arrayinitialsize);		\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISFULL_PROTOTYPE(name, prefix)				\
{									\
  return root->count == name##_gct_arraymaxsize;			\
}									\
/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */

#define _GCT_CONT_DARRAY_FOREACH_FROM(name, root, idx, ...)        \
 _GCT_CONT_ARRAYS_FOREACH_FROM(name, DARRAY, root, idx, __VA_ARGS__)

#define _GCT_CONT_DARRAY_FOREACH(name, root, ...)                  \
 _GCT_CONT_DARRAY_FOREACH_FROM(name, root, -1, __VA_ARGS__)

#define _GCT_CONT_DARRAY_FOREACH_REVERSE_FROM(name, root, idx, ...) \
 _GCT_CONT_ARRAYS_FOREACH_REVERSE_FROM(name, DARRAY, root, idx, __VA_ARGS__)

#define _GCT_CONT_DARRAY_FOREACH_REVERSE(name, root, ...)          \
 _GCT_CONT_DARRAY_FOREACH_REVERSE_FROM(name, root, (root)->count, __VA_ARGS__)

#define _GCT_CONT_DARRAY_FOREACH_UNORDERED(name, root, ...)        \
 _GCT_CONT_DARRAY_FOREACH(name, root, __VA_ARGS__)

/***********************************************************************
 *      key field based functions
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_DARRAY_FCNS(attr, name, prefix,			\
                                 lockname, keyfield, ...)		\
_GCT_CONT_LOOKUP_FCNS(attr, name, DARRAY, prefix,			\
                      lockname, keyfield);				\
/* backslash-region-end */


#endif

