/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_DBITMAP_H_
#define _GCT_DBITMAP_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_bitops.h>
#include <gct/_cont_bitmap.h>

#define _GCT_CONT_DBITMAP_ARRAY_BASED      0
#define _GCT_CONT_DBITMAP_POINTER_BASED    0
#define _GCT_CONT_DBITMAP_ORDERED          1
#define _GCT_CONT_DBITMAP_SORTED           0

/* backslash-region-begin */
#define _GCT_CONT_DBITMAP_TYPES(name, type, init_size, ...)		\
									\
struct                                  name##_bitmap_s			\
{									\
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;			\
  type          *data;							\
  size_t        size;							\
};									\
									\
typedef _gct_sindex_t                   name##_index_t;			\
typedef _gct_sindex_t                   name##_itembase_t;		\
typedef gct_bool_t                     name##_item_t;			\
typedef struct name##_bitmap_s          name##_root_t;			\
typedef type                            name##_entry_t;			\
									\
static const size_t							\
name##_gct_init_size = init_size;					\
									\
static const name##_index_t						\
name##_null_index = -1;							\
									\
typedef gct_bool_t _gct_##name##_keyfield_valuetype_t;			\
									\
/* backslash-region-end */

#define _GCT_CONT_DBITMAP_BITCOUNT(root, name) ((root)->size * sizeof (name##_entry_t) * 8)
#define _GCT_CONT_DBITMAP_BYTECOUNT(root, name) ((root)->size * sizeof (name##_entry_t))
#define _GCT_CONT_DBITMAP_WORDCOUNT(root, name) ((root)->size)

/* backslash-region-begin */
#define _GCT_CONT_DBITMAP_FCNS(attr, name, prefix, lockname, algo, ...)	\
 _GCT_CONT_BITMAPS_FCNS(attr, name, prefix, lockname, DBITMAP)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)				\
{									\
  size_t wc = _GCT_CONT_BITMAP_ALIGN_SIZE(name##_gct_init_size, name##_entry_t); \
									\
  if (name##_gct_init_size > 0						\
      && (root->data = calloc(wc, sizeof (name##_entry_t))) == NULL)	\
    return -1;								\
									\
  root->size = wc;							\
									\
  if (_GCT_LOCK_INIT(lockname, &root->lock))				\
    {									\
      free(root->data);							\
      return -1;							\
    }									\
									\
  return 0;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)				\
{									\
  free(root->data);							\
  _GCT_LOCK_DESTROY(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_RESIZE_PROTOTYPE(name, prefix)				\
{									\
  size_t wc = _GCT_CONT_BITMAP_ALIGN_SIZE(size, name##_entry_t);	\
  void *data;								\
  gct_err_t res = -1;							\
  unsigned int i;							\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  if ((data = realloc(root->data, wc * sizeof (name##_entry_t))) != NULL || wc == 0) \
    {									\
      root->data = data;						\
      res = 0;								\
									\
      for (i = root->size; i < wc; i++)					\
        root->data[i] = 0;						\
									\
      root->size = wc;							\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_KEY_DBITMAP_FCNS(attr, name, prefix, lockname, algo)	\
 _GCT_CONT_KEY_BITMAPS_FCNS(attr, name, prefix, lockname, DBITMAP)	\
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_DBITMAP_FOREACH(name, root, ...)			\
 _GCT_CONT_BITMAPS_FOREACH(name, root, DBITMAP, __VA_ARGS__)		\
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_DBITMAP_FOREACH_REVERSE(name, root, ...)		\
 _GCT_CONT_BITMAPS_FOREACH_REVERSE(name, root, DBITMAP, __VA_ARGS__)	\
/* backslash-region-end */

/* backslash-region-begin */
#define _GCT_CONT_DBITMAP_FOREACH_UNORDERED(name, root, ...)		\
 _GCT_CONT_BITMAPS_FOREACH(name, root, DBITMAP, __VA_ARGS__)		\
/* backslash-region-end */

#endif

