/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Double Linked List template.

    The list is implemented with head and tail pointers in container
    root. Null pointers are used as list terminators.
*/

#ifndef _GCT_CONT_DLIST_H_
#define _GCT_CONT_DLIST_H_

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_entry.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_pushpop_array.h>
#include <gct/_cont_lists.h>

#define _GCT_CONT_DLIST_ARRAY_BASED     0
#define _GCT_CONT_DLIST_POINTER_BASED   1
#define _GCT_CONT_DLIST_ORDERED         1
#define _GCT_CONT_DLIST_SORTED          0
#define _GCT_CONT_DLIST_REORDER         1

/***********************************************************************
 *      types
 */

typedef struct                          _gct_dlist_entry_s
{
  struct _gct_dlist_entry_s             *next;
  struct _gct_dlist_entry_s             *prev;
}                                       _gct_DLIST_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_DLIST_TYPES(name, type, entryfield)

typedef struct                          name##_root_s
{
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;
  struct _gct_dlist_entry_s             *head;
  struct _gct_dlist_entry_s             *tail;
}                                       name##_root_t;

typedef struct _gct_dlist_entry_s       name##_entry_t;
typedef type                            name##_itembase_t;
typedef name##_itembase_t *             name##_item_t;
typedef name##_itembase_t *             name##_index_t;

static const name##_item_t
name##_null_item = NULL;

static const name##_index_t
name##_null_index = NULL;

static const uintptr_t
name##_gct_offsetof = offsetof(name##_itembase_t, entryfield);

_GCT_CONT_ORPHAN_EMPTY_FCNS(DLIST, name, GCT_CONTAINER_ORPHAN_CHK_##name)
/* backslash-region-end */


/***********************************************************************
 *      orphan checking
 */

_GCT_INTERNAL void
_gct_orphan_DLIST__set(struct _gct_dlist_entry_s *e)
{
  e->next = e;
}

_GCT_INTERNAL gct_bool_t
_gct_orphan_DLIST__chk(struct _gct_dlist_entry_s *e)
{
  return e != e->next;
}

/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_ROOT_INITIALIZER(name, lockname)
{
  .lock = _GCT_LOCK_INITIALIZER(lockname),
  .head = NULL,
  .tail = NULL,
}
/* backslash-region-end */

#define _GCT_CONT_DLIST_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_DLIST_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

#define _GCT_CONT_DLIST_ISEND(root, entry)      ((entry) == NULL)
#define _GCT_CONT_DLIST_EMPTY(root)             ((root)->head = (root)->tail = NULL)

/* backslash-region-begin */
#define _GCT_CONT_DLIST_FCNS(attr, name, prefix, lockname, ...)

_GCT_CONT_LISTS_INIT_FCNS(attr, name, DLIST, prefix, lockname)

_GCT_CONT_LISTS_ROOT_FCNS(attr, name, DLIST, prefix, lockname, head)

_GCT_NONNULL(1, 2, 3)
attr GCT_CONT_SET_PROTOTYPE(name, prefix)
{
  gct_err_t          res = -1;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  {
    name##_entry_t *index_e = _GCT_CONT_GET_ENTRY(name, index);
    name##_entry_t *item_e = _GCT_CONT_GET_ENTRY(name, item);

    if (_GCT_CONT_ORPHAN_CHK(DLIST, name, index_e))
      {
        name##_entry_t *next;
        name##_entry_t *prev;

        _GCT_CONT_REFNEW(name, item);

        next = item_e->next = index_e->next;
        prev = item_e->prev = index_e->prev;
        *(prev ? &prev->next : &root->head) = item_e;
        *(next ? &next->prev : &root->tail) = item_e;

        _GCT_CONT_ORPHAN_SET(DLIST, name, index_e);
        _GCT_CONT_REFDROP(name, index);
        res = 0;
      }
  }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_CONT_LISTS_NEXT_HEAD_FCNS(attr, name, DLIST,
                               prefix, lockname,
                               NEXT, HEAD, next, head)

_GCT_CONT_LISTS_NEXT_HEAD_FCNS(attr, name, DLIST,
                               prefix, lockname,
                               PREV, TAIL, prev, tail)

_GCT_CONT_LISTS_COUNT_FCNS(attr, name, DLIST, prefix,
                           lockname, head)

_GCT_NONNULL(1, 2, 3)
attr GCT_CONT_INSERT_PRE_PROTOTYPE(name, prefix)
{
  name##_entry_t        *index_e;
  size_t                res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  index_e = _GCT_CONT_GET_ENTRY(name, index);

  if (_GCT_CONT_ORPHAN_CHK(DLIST, name, index_e))
    {
      name##_entry_t    *item_e;

      _GCT_CONT_REFNEW(name, item);

      item_e = _GCT_CONT_GET_ENTRY(name, item);
      item_e->next = index_e;
      item_e->prev = index_e->prev;
      *(index_e->prev ? &index_e->prev->next : &root->head) = item_e;
      index_e->prev = item_e;

      res = 1;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_NONNULL(1, 2, 3)
attr GCT_CONT_INSERT_POST_PROTOTYPE(name, prefix)
{
  name##_entry_t        *index_e;
  size_t                res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  index_e = _GCT_CONT_GET_ENTRY(name, index);

  if (_GCT_CONT_ORPHAN_CHK(DLIST, name, index_e))
    {
      name##_entry_t    *item_e;

      _GCT_CONT_REFNEW(name, item);

      item_e = _GCT_CONT_GET_ENTRY(name, item);
      item_e->prev = index_e;
      item_e->next = index_e->next;
      *(index_e->next ? &index_e->next->prev : &root->tail) = item_e;
      index_e->next = item_e;

      res = 1;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

_GCT_CONT_LISTS_CLEAR_FCNS(attr, name, DLIST, prefix, lockname, head)

_GCT_INTERNAL void
prefix##_remove_        (name##_root_t *root,
                         name##_entry_t *entry)
{
  name##_entry_t *next = entry->next;
  name##_entry_t *prev = entry->prev;

  *(prev ? &prev->next : &root->head) = next;
  *(next ? &next->prev : &root->tail) = prev;

  _GCT_CONT_ORPHAN_SET(DLIST, name, entry);
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix)
{
  gct_err_t  res = -1;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  {
    name##_entry_t      *index_e = _GCT_CONT_GET_ENTRY(name, index);

    if (_GCT_CONT_ORPHAN_CHK(DLIST, name, index_e))
      {
        prefix##_remove_(root, index_e);
        _GCT_CONT_REFDROP(name, index);

        res = 0;
      }

    _GCT_LOCK_UNLOCK(lockname, &root->lock);
  }

  return res;
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)
{
  name##_entry_t        *head;
  name##_entry_t        *item_e;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);
  _GCT_CONT_REFNEW(name, item);

  head = root->head;
  item_e = _GCT_CONT_GET_ENTRY(name, item);
  item_e->prev = 0;
  item_e->next = head;
  *(head ? &head->prev : &root->tail) = item_e;
  root->head = item_e;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return 1;
}

_GCT_NONNULL(1, 2)
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)
{
  name##_entry_t        *tail;
  name##_entry_t        *item_e;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);
  _GCT_CONT_REFNEW(name, item);

  tail = root->tail;
  item_e = _GCT_CONT_GET_ENTRY(name, item);

  item_e->next = 0;
  item_e->prev = tail;
  *(tail ? &tail->next : &root->head) = item_e;
  root->tail = item_e;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return 1;
}

_GCT_NONNULL(1)
attr GCT_CONT_POP_PROTOTYPE(name, prefix)
{
  name##_entry_t        *head;
  name##_item_t         item = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if ((head = root->head))
    {
      item = _GCT_CONT_GET_ITEM(name, head);
      prefix##_remove_(root, head);
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return item;
}

_GCT_NONNULL(1)
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix)
{
  name##_entry_t        *tail;
  name##_item_t         item = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if ((tail = root->tail))
    {
      item = _GCT_CONT_GET_ITEM(name, tail);
      prefix##_remove_(root, tail);
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return item;
}

_GCT_CONT_PUSHPOP_ARRAY_FCNS(attr, name, prefix, prefix, lockname);

_GCT_NONNULL(1, 2)
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)
_GCT_CONT_LISTS_FOREACH_FUNCCODE(attr, name, DLIST, prefix,
                                 lockname, head, next)

_GCT_NONNULL(1, 2)
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)
_GCT_CONT_LISTS_FOREACH_FUNCCODE(attr, name, DLIST, prefix,
                                 lockname, tail, prev)

_GCT_NONNULL(1)
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)
{
  name##_entry_t        *entry;
  gct_err_t          res = 0;

  _GCT_LOCK_RDLOCK(lockname, &root->lock);

  for (entry = root->head; !res && entry != NULL; entry = entry->next)
    {
      if (entry->next == entry || entry->prev == entry)
        res |= 1;

      if (root->head == entry && entry->prev != NULL)
        res |= 2;

      if (root->tail == entry && entry->next != NULL)
        res |= 4;

      if (entry->next != NULL && entry->next->prev != entry)
        res |= 8;

      if (entry->prev != NULL && entry->prev->next != entry)
        res |= 16;

      if (entry->next == root->head)
        res |= 32;

      if (entry->prev == root->tail)
        res |= 64;
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);

  return res;
}

/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_FOREACH(name, root, ...)
 _GCT_CONT_LISTS_FOREACH_BASE(DLIST, name, root,
                              (root)->head, next, __VA_ARGS__)
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_DLIST_FOREACH_FROM(name, root, index, ...)
 _GCT_CONT_LISTS_FOREACH_BASE(DLIST, name, root,
			      _GCT_CONT_GET_ENTRY(name, index)->next, next, __VA_ARGS__)
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_DLIST_FOREACH_REVERSE(name, root, ...)
 _GCT_CONT_LISTS_FOREACH_BASE(DLIST, name, root,
                              (root)->tail, prev, __VA_ARGS__)
/* backslash-region-end */


/* backslash-region-begin */
#define _GCT_CONT_DLIST_FOREACH_REVERSE_FROM(name, root, index, ...)
 _GCT_CONT_LISTS_FOREACH_BASE(DLIST, name, root,
			      _GCT_CONT_GET_ENTRY(name, index)->prev, prev, __VA_ARGS__)
/* backslash-region-end */


#define _GCT_CONT_DLIST_FOREACH_UNORDERED(name, root, ...)         \
 _GCT_CONT_DLIST_FOREACH(name, root, __VA_ARGS__)


/***********************************************************************
 *      key field based functions
 */

_GCT_CONT_LISTS_SORT_FCNS(DLIST, _gct_DLIST_entry_t)




/*
 * DLIST links rebuild after merge/sort template code
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_REBUILD_CODE(name, lockname, cmp, entry, root)
{
  name##_entry_t *prev;
  root->head = entry;

  for (prev = NULL; entry != NULL; entry = entry->next)
    {
      entry->prev = prev;
      prev = entry;
    }

  root->tail = prev;
}
/* backslash-region-end */





/*
 * DLIST sort glue function template, call sort and rebuild prev links
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_SORT_FUNCCODE(name, lockname, cmp)
{
  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  if (root->head != NULL && root->head->next != NULL)
    {
      name##_entry_t *tmp = _gct_DLIST_sort(root->head, cmp);

      _GCT_CONT_DLIST_REBUILD_CODE(name, lockname, cmp, tmp, root);
    }

  _GCT_LOCK_UNLOCK(lockname, &root->lock);
}
/* backslash-region-end */





/*
 * DLIST merge glue function template, call merge and rebuild prev links
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_MERGE_FUNCCODE(name, lockname, cmp)
{
  _GCT_LOCK_WRLOCK(lockname, &src->lock);
  _GCT_LOCK_WRLOCK(lockname, &dest->lock);

  if (src->head != NULL)
    {
      if (dest->head == NULL)
	{
	  /* move from src to dest */
	  dest->head = src->head;
	  dest->tail = src->tail;
	}
      else
	{
	  name##_entry_t *tmp = _gct_DLIST_merge(dest->head, src->head, cmp);

	  _GCT_CONT_DLIST_REBUILD_CODE(name, lockname, cmp, tmp, dest);
	}

      _GCT_CONT_DLIST_EMPTY(src);
    }

  _GCT_LOCK_UNLOCK(lockname, &dest->lock);
  _GCT_LOCK_UNLOCK(lockname, &src->lock);
}
/* backslash-region-end */





/*
 * DLIST Ordered insertion function template
 */

/* backslash-region-begin */
#define _GCT_CONT_DLIST_ORDERED_INSERT_FUNCCODE(name, lockname, cmp)
{
  name##_entry_t        *index_e, *item_e, **prev;
  size_t                res = 0;

  _GCT_LOCK_WRLOCK(lockname, &root->lock);

  item_e = _GCT_CONT_GET_ENTRY(name, item);

  index_e = NULL;
  for (prev = &root->head; *prev != NULL; (index_e = *prev), (prev = &(*prev)->next))
    if (cmp(item_e, *prev))
      break;

  _GCT_CONT_REFNEW(name, item);

  item_e->next = *prev;
  item_e->prev = index_e;
  *(*prev ? &(*prev)->prev : &root->tail) = item_e;
  *prev = item_e;

  _GCT_LOCK_UNLOCK(lockname, &root->lock);
}
/* backslash-region-end */





/* backslash-region-begin */
#define _GCT_CONT_KEY_DLIST_FCNS(attr, name, prefix,
                                 lockname, keyfield, ...)

_GCT_CONT_LOOKUP_FCNS(attr, name, DLIST, prefix, lockname, keyfield);

_GCT_CONT_LISTS_CMP_FCNS(name, prefix, DLIST, keyfield)

attr GCT_CONT_SORT_ASCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_SORT_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend)

attr GCT_CONT_SORT_DESCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_SORT_FUNCCODE(name, lockname, prefix##_gct_cmp_descend)

attr GCT_CONT_SORT_STABLE_ASCEND_PROTOTYPE(name, prefix, keyfield)
{
  prefix##_sort_ascend(root);
}

attr GCT_CONT_SORT_STABLE_DESCEND_PROTOTYPE(name, prefix, keyfield)
{
  prefix##_sort_descend(root);
}

attr GCT_CONT_SORTED_ASCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_LISTS_SORTED_FUNCCODE(DLIST, name, lockname, prefix##_gct_cmp_ascend, root->head)

attr GCT_CONT_SORTED_DESCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_LISTS_SORTED_FUNCCODE(DLIST, name, lockname, prefix##_gct_cmp_descend, root->head)

attr GCT_CONT_INSERT_ASCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_ORDERED_INSERT_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend)

attr GCT_CONT_INSERT_DESCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_ORDERED_INSERT_FUNCCODE(name, lockname, prefix##_gct_cmp_descend)

attr GCT_CONT_MERGE_ASCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_MERGE_FUNCCODE(name, lockname, prefix##_gct_cmp_ascend)

attr GCT_CONT_MERGE_DESCEND_PROTOTYPE(name, prefix, keyfield)
_GCT_CONT_DLIST_MERGE_FUNCCODE(name, lockname, prefix##_gct_cmp_descend)
/* backslash-region-end */



#endif

