/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

    Linked list (same as BLIST) with hash table on list head.
*/

#ifndef _GCT_CONT_HASHLIST_H_
#define _GCT_CONT_HASHLIST_H_

#include <stddef.h>

#include <gct/_cont_access.h>
#include <gct/_cont_keyfield.h>
#include <gct/_cont_lookup.h>
#include <gct/_cont_entry.h>
#include <gct/_cont_pushpop_array.h>

#define _GCT_CONT_HASHLIST_ARRAY_BASED    0
#define _GCT_CONT_HASHLIST_POINTER_BASED  1
#define _GCT_CONT_HASHLIST_ORDERED        0
#define _GCT_CONT_HASHLIST_SORTED         0

/***********************************************************************
 *      types
 */

typedef struct                          _gct_hashlist_entry_s
{
  struct _gct_hashlist_entry_s          *next;
  struct _gct_hashlist_entry_s          **prev;
}                                       _gct_HASHLIST_entry_t;

/* backslash-region-begin */
#define _GCT_CONT_HASHLIST_TYPES(name, type, entryfield, size)		\
									\
_GCT_CONT_COUNTER_TYPES(name)						\
									\
typedef struct                          name##_root_s			\
{									\
  struct _gct_hashlist_entry_s          *head[size];			\
  _GCT_CONT_COUNTER_T(name)             counter;			\
  _GCT_LOCK_TYPES(GCT_CONTAINER_LOCK_##name) lock;			\
}                                       name##_root_t;			\
									\
typedef struct _gct_hashlist_entry_s    name##_entry_t;			\
typedef type                            name##_itembase_t;		\
typedef name##_itembase_t *             name##_item_t;			\
typedef name##_itembase_t *             name##_index_t;			\
									\
static const name##_item_t						\
name##_null_item = NULL;						\
									\
static const name##_index_t						\
name##_null_index = NULL;						\
									\
static const uintptr_t							\
name##_gct_offsetof = offsetof(name##_itembase_t, entryfield);		\
									\
static const size_t							\
name##_gct_hashsize = (size);						\
									\
_GCT_CONT_COUNTER_FCNS(name)						\
_GCT_CONT_ORPHAN_EMPTY_FCNS(HASHLIST, name, GCT_CONTAINER_ORPHAN_CHK_##name) \
/* backslash-region-end */


/***********************************************************************
 *      orphan checking
 */

_GCT_INTERNAL void
_gct_orphan_HASHLIST__set(struct _gct_hashlist_entry_s *e)
{
  e->next = e;
}

_GCT_INTERNAL gct_bool_t
_gct_orphan_HASHLIST__chk(struct _gct_hashlist_entry_s *e)
{
  return e != e->next;
}


/***********************************************************************
 *      intializer
 */

/* backslash-region-begin */
#define _GCT_CONT_HASHLIST_ROOT_INITIALIZER(name, lockname)		\
{									\
  .lock = _GCT_LOCK_INITIALIZER(lockname),				\
  .head = { 0 },							\
}									\
/* backslash-region-end */

#define _GCT_CONT_HASHLIST_ROOT_DECLARATOR(name, lockname, symbol) \
 name##_root_t symbol = _GCT_CONT_HASHLIST_ROOT_INITIALIZER(name, lockname)

/***********************************************************************
 *      access functions
 */

/* backslash-region-begin */
#define _GCT_CONT_HASHLIST_FCNS(attr, name, prefix, lockname, keyfield)	\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISORPHAN_PROTOTYPE(name, prefix)				\
{									\
  return !_GCT_CONT_ORPHAN_CHK(HASHLIST, name, _GCT_CONT_GET_ENTRY(name, item)); \
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ORPHAN_PROTOTYPE(name, prefix)				\
{									\
  _GCT_CONT_ORPHAN_SET(HASHLIST, name, _GCT_CONT_GET_ENTRY(name, item)); \
}									\
									\
attr GCT_CONT_ISNULL_PROTOTYPE(name, prefix)				\
{									\
  return !index;							\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISEMPTY_PROTOTYPE(name, prefix)				\
{									\
  gct_bool_t           res;						\
  uintptr_t             i;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  if (_GCT_CONT_COUNTER_ENABLED(name))					\
    res = _GCT_CONT_COUNTER_GET(name, root) == 0;			\
  else									\
    {									\
      res = 1;								\
      for (i = 0; i < name##_gct_hashsize; i++)				\
        if (root->head[i])						\
          {								\
            res = 0;							\
            break;							\
          }								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_ISFULL_PROTOTYPE(name, prefix)				\
{									\
  return 0;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_GET_PROTOTYPE(name, prefix)				\
{									\
  return index;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETINDEX_PROTOTYPE(name, prefix)				\
{									\
  return ptr;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_GETPTR_PROTOTYPE(name, prefix)				\
{									\
  return index;								\
}									\
									\
_GCT_DEPRECATED								\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_NEXT_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                item;					\
  name##_entry_t                *entry;					\
  uintptr_t                     i;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  if (!(entry = _GCT_CONT_GET_ENTRY(name, index)->next))		\
    for (i = (_GCT_CONT_HASH_FIELD(name, keyfield, index)		\
              % name##_gct_hashsize) + 1;				\
         !entry && i < name##_gct_hashsize; i++)			\
        entry = root->head[i];						\
									\
  item = entry								\
       ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, entry))	\
       : NULL;								\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_DEPRECATED								\
_GCT_NONNULL(1)								\
attr GCT_CONT_HEAD_PROTOTYPE(name, prefix)				\
{									\
  name##_index_t                item;					\
  name##_entry_t                *entry = 0;				\
  uintptr_t                     i;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = 0; !entry && i < name##_gct_hashsize; i++)			\
    entry = root->head[i];						\
									\
  item = entry								\
       ? _GCT_CONT_REFNEW(name, _GCT_CONT_GET_ITEM(name, entry))	\
       : NULL;								\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_COUNT_PROTOTYPE(name, prefix)				\
{									\
  uintptr_t             i;						\
  size_t                s;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  name##_entry_t        *entry;						\
									\
  if (_GCT_CONT_COUNTER_ENABLED(name))					\
    s = _GCT_CONT_COUNTER_GET(name, root);				\
  else									\
    for (s = i = 0; i < name##_gct_hashsize; i++)			\
      for (entry = root->head[i]; entry; s++)				\
        entry = entry->next;						\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return s;								\
}									\
									\
attr GCT_CONT_SIZE_PROTOTYPE(name, prefix)				\
{									\
  return name##_gct_hashsize;						\
}									\
									\
_GCT_INTERNAL								\
_GCT_NONNULL(1)								\
void prefix##_remove_   (name##_entry_t *entry)				\
{									\
  name##_entry_t *next = entry->next;					\
  name##_entry_t **prev = entry->prev;					\
									\
  if ((*prev = next))							\
    next->prev = prev;							\
									\
  _GCT_CONT_ORPHAN_SET(HASHLIST, name, entry);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CLEAR_PROTOTYPE(name, prefix)				\
{									\
  uintptr_t             i;						\
  name##_entry_t        *entry, *next;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = 0; i < name##_gct_hashsize; i++)				\
    {									\
      for (entry = root->head[i]; entry; entry = next)			\
        {								\
          _GCT_CONT_ORPHAN_SET(HASHLIST, name, entry);			\
          next = entry->next;						\
          _GCT_CONT_REFDROP(name, _GCT_CONT_GET_ITEM(name, entry));	\
        }								\
      root->head[i] = NULL;						\
    }									\
									\
  _GCT_CONT_COUNTER_SET(name, root, 0);					\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_REMOVE_PROTOTYPE(name, prefix)				\
{									\
  gct_err_t          res = -1;						\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
									\
  {									\
    name##_entry_t      *index_e = _GCT_CONT_GET_ENTRY(name, index);	\
									\
    if (_GCT_CONT_ORPHAN_CHK(HASHLIST, name, index_e))			\
      {									\
        prefix##_remove_(index_e);					\
        _GCT_CONT_REFDROP(name, index);					\
        _GCT_CONT_COUNTER_ADD(name, root, -1);				\
        res = 0;							\
      }									\
  }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_PUSH_PROTOTYPE(name, prefix)				\
{									\
  name##_entry_t        **entry, *entry_, *item_e;			\
									\
  _GCT_LOCK_WRLOCK(lockname, &root->lock);				\
  _GCT_CONT_REFNEW(name, item);						\
									\
  entry = &root->head[_GCT_CONT_HASH_FIELD(name, keyfield, item)	\
                      % name##_gct_hashsize];				\
  entry_ = *entry;							\
  item_e = _GCT_CONT_GET_ENTRY(name, item);				\
									\
  item_e->prev = entry;							\
  item_e->next = entry_;						\
  if (entry_ != NULL)							\
    entry_->prev = &item_e->next;					\
  *entry = item_e;							\
  _GCT_CONT_COUNTER_ADD(name, root, 1);					\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return 1;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_PUSHBACK_PROTOTYPE(name, prefix)				\
{									\
  return prefix##_push(root, item);					\
}									\
									\
_GCT_INTERNAL								\
_GCT_NONNULL(1)								\
GCT_CONT_POP_PROTOTYPE(name, prefix##_gct)				\
{									\
  name##_index_t                item = 0;				\
  name##_entry_t                *entry = 0;				\
  uintptr_t                     i;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = 0; i < name##_gct_hashsize; i++)				\
    if ((entry = root->head[i]))					\
      break;								\
									\
  if (entry)								\
    {									\
      item = _GCT_CONT_GET_ITEM(name, entry);				\
      prefix##_remove_(entry);						\
      _GCT_CONT_COUNTER_ADD(name, root, -1);				\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return item;								\
}									\
									\
_GCT_NONNULL(1)								\
_GCT_DEPRECATED								\
attr GCT_CONT_POP_PROTOTYPE(name, prefix)				\
{									\
  return prefix##_gct_pop(root);					\
}									\
									\
_GCT_NONNULL(1)								\
_GCT_DEPRECATED								\
attr GCT_CONT_POPBACK_PROTOTYPE(name, prefix)				\
{									\
  return prefix##_gct_pop(root);					\
}									\
									\
attr GCT_CONT_PUSH_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  uintptr_t     i;							\
									\
  for (i = 0; i < size; i++)						\
    prefix##_push(root, item[i]);					\
									\
  return size;								\
}									\
									\
_GCT_DEPRECATED								\
attr GCT_CONT_POP_ARRAY_PROTOTYPE(name, prefix)				\
{									\
  uintptr_t     i;							\
									\
  for (i = 0; i < size; i++)						\
    item[i] = prefix##_gct_pop(root);					\
									\
  return size;								\
}									\
									\
attr GCT_CONT_PUSHBACK_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  uintptr_t     i;							\
									\
  for (i = 0; i < size; i++)						\
    prefix##_push(root, item[i]);					\
									\
  return size;								\
}									\
									\
_GCT_DEPRECATED								\
attr GCT_CONT_POPBACK_ARRAY_PROTOTYPE(name, prefix)			\
{									\
  uintptr_t     i;							\
									\
  for (i = 0; i < size; i++)						\
    item[i] = prefix##_gct_pop(root);					\
									\
  return size;								\
}									\
									\
_GCT_INTERNAL intptr_t							\
prefix##_foreach_(name##_root_t *root,					\
                  GCT_CONTAINER_ITERATOR(name, * const fcn),		\
                  va_list ap)						\
{									\
  va_list               aq;						\
  uintptr_t             i;						\
  name##_entry_t        *entry;						\
  intptr_t              res = 0;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = 0; !res && i < name##_gct_hashsize; i++)			\
    for (entry = root->head[i]; !res && entry; entry = entry->next)	\
      {									\
        va_copy(aq, ap);						\
        res = fcn(_GCT_CONT_GET_ITEM(name, entry), aq);			\
        va_end(aq);							\
      }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_PROTOTYPE(name, prefix)				\
{									\
  intptr_t      res;							\
  va_list       ap;							\
									\
  va_start(ap, fcn);							\
  res = prefix##_foreach_(root, fcn, ap);				\
  va_end(ap);								\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1, 2)							\
attr GCT_CONT_FOREACH_REVERSE_PROTOTYPE(name, prefix)			\
{									\
  intptr_t      res;							\
  va_list       ap;							\
									\
  va_start(ap, fcn);							\
  res = prefix##_foreach_(root, fcn, ap);				\
  va_end(ap);								\
									\
  return res;								\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_INIT_PROTOTYPE(name, prefix)				\
{									\
  uintptr_t             i;						\
									\
  for (i = 0; i < name##_gct_hashsize; i++)				\
    root->head[i] = NULL;						\
									\
  _GCT_CONT_COUNTER_SET(name, root, 0);					\
									\
  return _GCT_LOCK_INIT(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_DESTROY_PROTOTYPE(name, prefix)				\
{									\
  _GCT_LOCK_DESTROY(lockname, &root->lock);				\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_CHECK_PROTOTYPE(name, prefix)				\
{									\
  uintptr_t             i;						\
  name##_entry_t        *entry;						\
  gct_err_t          res = 0;						\
  size_t                count = 0;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (i = 0; !res && i < name##_gct_hashsize; i++)			\
    for (entry = root->head[i]; !res && entry != NULL; entry = entry->next) \
      {									\
        if (entry->next == entry)					\
          res |= 1;							\
									\
        if (root->head[i] == entry && entry->prev != &root->head[i])	\
          res |= 2;							\
									\
        if (entry->next != NULL && entry->next->prev != &entry->next)	\
          res |= 4;							\
									\
        if (entry->next == root->head[i])				\
          res |= 8;							\
									\
        count++;							\
      }									\
									\
  if (!res && _GCT_CONT_COUNTER_ENABLED(name))				\
    if (_GCT_CONT_COUNTER_GET(name, root) != count)			\
      res |= 16;							\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return res;								\
}									\
/* backslash-region-end */




/***********************************************************************
 *      iteration macros
 */


/* backslash-region-begin */
#define _GCT_CONT_HASHLIST_FOREACH(name, root, ...)			\
  _GCT_CONT_LABEL_DECL(_gct_break);					\
  uintptr_t             _gct_i;						\
  name##_entry_t        *_gct_entry;					\
  name##_entry_t        *_gct_next;					\
									\
  for (_gct_i = 0; _gct_i < name##_gct_hashsize; _gct_i++)		\
    for (_gct_entry = (root)->head[_gct_i];				\
         _gct_entry; _gct_entry = _gct_next)				\
      {									\
        _GCT_CONT_LABEL_DECL(_gct_continue);				\
        name##_item_t   item = _GCT_CONT_GET_ITEM(name, _gct_entry);	\
        _GCT_UNUSED name##_index_t      index = item;			\
        _gct_next = _gct_entry->next;					\
									\
        { __VA_ARGS__ }							\
        _GCT_CONT_LABEL(_gct_continue);					\
      }									\
  _GCT_CONT_LABEL(_gct_break);						\
/* backslash-region-end */


#define _GCT_CONT_HASHLIST_FOREACH_REVERSE(name, root, ...)        \
 _GCT_CONT_HASHLIST_FOREACH(name, root, __VA_ARGS__)

#define _GCT_CONT_HASHLIST_FOREACH_UNORDERED(name, root, ...)      \
 _GCT_CONT_HASHLIST_FOREACH(name, root, __VA_ARGS__)

#define _GCT_CONT_HASHLIST_FOREACH_FROM(name, root, index, ...)    \
 _GCT_ERROR_NA(name);

#define _GCT_CONT_HASHLIST_FOREACH_REVERSE_FROM(name, root, index, ...) \
 _GCT_ERROR_NA(name);


/***********************************************************************
 *      key field based functions
 */

/* backslash-region-begin */
#define _GCT_CONT_KEY_HASHLIST_FCNS(attr, name, prefix,			\
                                    lockname, keyfield, ...)		\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_LOOKUP_PROTOTYPE(name, prefix, keyfield)			\
{									\
  _gct_key_hash_t       i;						\
  name##_entry_t        *entry;						\
  name##_item_t         item = NULL;					\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  i = _GCT_CONT_HASH_ARG(name, keyfield, value)				\
    % name##_gct_hashsize;						\
									\
  for (entry = root->head[i]; entry; entry = entry->next)		\
    {									\
      item = _GCT_CONT_GET_ITEM(name, entry);				\
									\
      if (!_GCT_CONT_COMPARE_FIELD_ARG(name, keyfield, item, value))	\
        {								\
          item = _GCT_CONT_REFNEW(name, item);				\
          break;							\
        }								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return entry ? item : NULL;						\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_LOOKUP_LAST_PROTOTYPE(name, prefix, keyfield)		\
{									\
  return prefix##_lookup(root, value);					\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_LOOKUP_NEXT_PROTOTYPE(name, prefix, keyfield)		\
{									\
  name##_entry_t        *entry;						\
  name##_item_t         item;						\
									\
  _GCT_LOCK_RDLOCK(lockname, &root->lock);				\
									\
  for (entry = _GCT_CONT_GET_ENTRY(name, index)->next;			\
       entry; entry = entry->next)					\
    {									\
      item = _GCT_CONT_GET_ITEM(name, entry);				\
									\
      if (!_GCT_CONT_COMPARE_FIELD_ARG(name, keyfield, item, value))	\
        {								\
          item = _GCT_CONT_REFNEW(name, item);				\
          break;							\
        }								\
    }									\
									\
  _GCT_LOCK_UNLOCK(lockname, &root->lock);				\
									\
  return entry ? item : NULL;						\
}									\
									\
_GCT_NONNULL(1)								\
attr GCT_CONT_LOOKUP_PREV_PROTOTYPE(name, prefix, keyfield)		\
{									\
  return prefix##_lookup_next(root, index, value);			\
}									\
									\
_GCT_CONT_REMOVE_KEY_FCNS(attr, name, HASHLIST, prefix,			\
                               lockname, keyfield)			\
/* backslash-region-end */



#endif

