/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_OBJECT_REFCOUNT_H_
#define _GCT_OBJECT_REFCOUNT_H_

#include <stdarg.h>

#include <gct/_platform.h>
#include <gct/_atomic.h>
#include <gct/_object.h>
#include <gct/_helpers.h>

#define _GCT_OBJ_REFCOUNT_ENTRY_FIELDS(name) \
 _gct_atomic_t           refcnt;



/**
   Object manager with simple reference counting. the delete()
   function is not available here.
 */

/* backslash-region-begin */
#define _GCT_OBJ_REFCOUNT_FCNS(attr, name, prefix, f)			\
									\
attr _GCT_OBJ_CONSTRUCT_PROTOTYPE(name, prefix)				\
{									\
  name##_object_t obj = storage;					\
  gct_err_t err;							\
									\
  _GCT_ASSERT(storage != NULL);						\
  _gct_atomic_init(&obj->f.refcnt);					\
  _gct_atomic_set(&obj->f.refcnt, 1);					\
  obj->f.storage_free = sfree;						\
									\
  err = name##_constructor(obj, ap);					\
									\
  /* prevent memset(obj, 0, sizeof(*obj)) in constructor */		\
  _GCT_ASSERT(_gct_atomic_get(&obj->f.refcnt) >= 1);			\
  _GCT_ASSERT(obj->f.storage_free == sfree);				\
									\
  if (err && sfree != NULL)						\
    sfree(storage);							\
									\
  return err;								\
}									\
									\
attr _GCT_OBJ_NEW_PROTOTYPE(name, prefix)				\
{									\
  name##_object_t obj = storage;					\
  object_storage_free_t *sfree = NULL;					\
  va_list ap;								\
									\
  if (obj == NULL)							\
    {									\
      obj = _gct_malloc(sizeof(*obj));					\
      sfree = &_gct_free;						\
    }									\
									\
  if (obj != NULL)							\
    {									\
      va_start(ap, storage);						\
      if (prefix##_construct(obj, sfree, ap))				\
        return NULL;							\
      va_end(ap);							\
    }									\
									\
  return obj;								\
}									\
									\
attr _GCT_OBJ_DELETE_PROTOTYPE(name, prefix)				\
{									\
  _GCT_STR_ERROR("delete() is illegal with reference counting");	\
  return 0;								\
}									\
									\
attr _GCT_OBJ_REFNEW_PROTOTYPE(name, prefix)				\
{									\
  _GCT_ASSERT(_gct_atomic_get(&object->f.refcnt) >= 0);			\
									\
  _gct_atomic_inc(&object->f.refcnt);					\
									\
  return object;							\
}									\
									\
attr _GCT_OBJ_REFCOUNT_PROTOTYPE(name, prefix)				\
{									\
  return _gct_atomic_get(&object->f.refcnt);				\
}									\
									\
attr _GCT_OBJ_REFDROP_PROTOTYPE(name, prefix)				\
{									\
  _GCT_ASSERT(_gct_atomic_get(&object->f.refcnt) > 0);			\
									\
  if (_gct_atomic_dec(&object->f.refcnt))				\
    return;								\
									\
  name##_destructor(object);						\
									\
  /* give a chance to destructor to abort */				\
  if (_gct_atomic_get(&object->f.refcnt) > 0)				\
    return;								\
									\
  _gct_atomic_destroy(&object->f.refcnt);				\
  if (object->f.storage_free != NULL)					\
    object->f.storage_free(object);					\
}									\
/* backslash-region-end */



/* backslash-region-begin */
#define _GCT_OBJ_REFCOUNT_INITIALIZER					\
{									\
  .storage_free = NULL,							\
  .refcnt = _GCT_ATOMIC_INITIALIZER(1)					\
}									\
/* backslash-region-end */

#endif

