/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#include "testsuite.h"

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define GCT_CONTAINER_LOCK_test TEST
#include "lock_test.h"


#include <gct/cont_clist.h>

struct aggregate_s
{
  long int				a;
  long int				b;
};

struct test_object_s
{
  _GCT_CONTAINER_ENTRY_TYPE(CLIST)	list_entry;
  int					id;
  char					*name;
  uint8_t				blob[14];
  unsigned int				mask;
  struct aggregate_s			agg;
};

GCT_CONTAINER_TYPES		(test, CLIST, struct test_object_s, list_entry);

GCT_CONTAINER_KEY_TYPES	(test, PTR, SCALAR, id, TESTKEYTYPE_EXTRA);
GCT_CONTAINER_KEY_PROTOTYPES	(test, static, list, id);

GCT_CONTAINER_PROTOTYPES	(test, static, list);
GCT_CONTAINER_FCNS		(test, CLIST, static, list, id);

GCT_CONTAINER_KEY_FCNS	(test, CLIST, static, list, id, TESTKEYFUNC_EXTRA);

unsigned int		i;

static GCT_CONTAINER_ITERATOR(test, iter_func)
{
  i = i * item->id + item->id;  
  return 0;
}

int main()
{
  test_root_t	list;
  struct test_object_s	obj[] =
    {
      {
	.id = 3,
	.name = "foo",
	.blob = "foo",
	.agg = { .a = 3 }
      },
      {
	.id = 5,
	.name = "bar",
	.blob = "bar",
	.agg = { .a = 5 }
      },
      {
	.id = 7,
	.name = "prettylongname",
	.blob = "prettylongname",
	.agg = { .a = 7 }
      },
      {
	.id = 11,
	.name = "test",
	.blob = "test",
	.agg = { .a = 11 }
      },
      {
	.id = 13,
	.name = "foo",
	.blob = "foo",
	.agg = { .a = 13 }
      },
#define ENTRY_COUNT	5

      {
	.id = 0,
	.name = "min",
	.blob = "min",
	.agg = { .a = 0 }
      },
      {
	.id = 30,
	.name = "max",
	.blob = "max",
	.agg = { .a = 30 }
      },
      {
	.id = 7,
	.name = "double",
	.blob = "double",
	.agg = { .a = 7 }
      },
      {
	.id = 10,
	.name = "any",
	.blob = "any",
	.agg = { .a = 10 }
      },

      {
	.id = -1,
	.name = "MIN",
	.blob = "MIN",
	.agg = { .a = -1 }
      },
      {
	.id = 12,
	.name = "MAX",
	.blob = "MAX",
	.agg = { .a = 12 }
      },
      {
	.id = 20,
	.name = "DOUBLE",
	.blob = "DOUBLE",
	.agg = { .a = 20 }
      },
      {
	.id = 32,
	.name = "ANY",
	.blob = "ANY",
	.agg = { .a = 32 }
      },
    };

#if GCT_CONTAINER_FEATURE(CLIST, ARRAY_BASED)
  static const int idx[ENTRY_COUNT] = { 0, 1, 1, 2, 2 };
#else
  struct test_object_s *idx[ENTRY_COUNT] = { &obj[0], &obj[0], &obj[0], &obj[0], &obj[0] };
#endif

  puts("Container ordered insertion test (" NAME_STR(CLIST) ", " NAME_STR(SCALAR)")");

  list_init(&list);

  /* push entries */

  assert(list_push(&list, &obj[0]) == 1);
  assert(!list_check(&list));
  /* 0 */

  assert(list_insert_pre(&list, idx[0], &obj[1]) == 1);
  assert(!list_check(&list));
  /* 1 0 */

  assert(list_insert_post(&list, idx[1], &obj[2]) == 1);
  assert(!list_check(&list));
  /* 1 0 2 */

  assert(list_insert_pre(&list, idx[2], &obj[3]) == 1);
  assert(!list_check(&list));
  /* 1 3 0 2 */

  assert(list_insert_post(&list, idx[3], &obj[4]) == 1);
  assert(!list_check(&list));
  /* 1 3 0 4 2 */


  i = 1;
  GCT_CONTAINER_FOREACH(test, CLIST, &list,
  {
    i = i * item->id + item->id;
  });

  assert(!list_check(&list));
  assert(i == 33404);


  i = 1;
  list_foreach(&list, iter_func);
  assert(i == 33404);


  i = 1;
  GCT_CONTAINER_FOREACH_REVERSE(test, CLIST, &list,
  {
    i = i * item->id + item->id;
  });
  assert(i == 32400);


  i = 1;
  list_foreach_reverse(&list, iter_func);
  assert(i == 32400);

#if GCT_CONTAINER_FEATURE(CLIST, REORDER)
  puts("Container sorting test (" NAME_STR(CLIST) ", " NAME_STR(SCALAR)")");

  list_sort_ascend(&list);
  assert(!list_check(&list));
  assert(list_issorted_ascend(&list));
  assert(!list_issorted_descend(&list));

  for (i = 5; i <= 8; i++)
    {
      list_insert_ascend(&list, &obj[i]);
      assert(!list_check(&list));
      assert(list_issorted_ascend(&list));
      assert(!list_issorted_descend(&list));
    }
  assert(list_count(&list) == 9);

  list_sort_descend(&list);
  assert(!list_check(&list));
  assert(!list_issorted_ascend(&list));
  assert(list_issorted_descend(&list));

  for (i = 9; i <= 12; i++)
    {
      list_insert_descend(&list, &obj[i]);
      assert(!list_check(&list));
      assert(!list_issorted_ascend(&list));
      assert(list_issorted_descend(&list));
    }
  assert(list_count(&list) == 13);

  list_clear(&list);
  test_root_t	list2;
  list_init(&list2);

  list_sort_descend(&list2);
  list_sort_ascend(&list2);

  for (i = 0; i <= 8; i++)
    {
      list_insert_ascend(&list, &obj[i]);
      assert(!list_check(&list));
    }
  assert(list_issorted_ascend(&list));
  assert(!list_issorted_descend(&list));

  list_merge_ascend(&list2, &list);
  assert(!list_check(&list2));
  assert(!list_check(&list));
  assert(list_count(&list) == 0);
  assert(list_count(&list2) == 9);
  assert(list_issorted_ascend(&list2));
  assert(!list_issorted_descend(&list2));

  list_sort_descend(&list2);
  assert(list_count(&list2) == 9);

  for (i = 9; i <= 12; i++)
    {
      list_insert_descend(&list, &obj[i]);
      assert(!list_check(&list));
    }
  assert(!list_issorted_ascend(&list));
  assert(list_issorted_descend(&list));
  assert(list_count(&list) == 4);

  list_merge_descend(&list2, &list);
  assert(!list_check(&list2));
  assert(!list_check(&list));
  assert(list_count(&list2) == 13);
  assert(list_count(&list) == 0);
  assert(list_issorted_descend(&list2));

  list_destroy(&list2);

#endif

  /* Check locking effectiveness */

#ifdef GCT_CONTAINER_LOCK_test
  puts("Container locking test (" NAME_STR(CLIST) ")");

  assert(list.lock == 0);
  list_wrlock(&list);
  assert(list.lock == 1);
  list_unlock(&list);
  assert(list.lock == 0);
  list_rdlock(&list);
  assert(list.lock == 2);
  list_unlock(&list);
  assert(list.lock == 0);

#endif

  list_destroy(&list);

  return 0;
}

